#!/usr/bin/env python
""" This example shows all mesh objects that can be created out of the box.
It also shows the different techniques to apply color to the meshes using
plain color, colormaps and texture.

On the website, this example also demonstrates the fly camera to fly through
the mesh objects.

"""

import numpy as np
import visvis as vv
from visvis import Point, Pointset
vv.figure()
a = vv.gca()

# Define points for the line
pp = Pointset(3)
pp.append(0,0,0); pp.append(0,1,0); pp.append(1,2,0); pp.append(0,2,1)

# Create all solids
box = vv.solidBox((0,0,0))
sphere = vv.solidSphere((3,0,0))
cone = vv.solidCone((6,0,0))
pyramid = vv.solidCone((9,0,0), N=4) # a cone with 4 faces is a pyramid
cylinder = vv.solidCylinder((0,3,0),(1,1,2))
ring = vv.solidRing((3,3,0))
teapot = vv.solidTeapot((6,3,0))
line = vv.solidLine(pp+Point(9,3,0), radius = 0.2)

# Let's put a face on that cylinder
# This works because 2D texture coordinates are automatically generated for
# the sphere, cone, cylinder and ring.
im = vv.imread('astronaut.png')
cylinder.SetTexture(im)

# Make the ring green
ring.faceColor = 'g'

# Make the sphere dull
sphere.specular = 0
sphere.diffuse = 0.4

# Show lines in yellow pyramid
pyramid.faceColor = 'y'
pyramid.edgeShading = 'plain'

# Colormap example
N = cone._vertices.shape[0]
cone.SetValues( np.linspace(0,1,N) )
cone.colormap = vv.CM_JET

# Show title and enter main loop
vv.title('All mesh objects that can be created out of the box')
app = vv.use()
app.Run()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" Module baseTexture

Defines the base TextureObject class (which is not a wobject or wibject),
and the colormap class that derives from it. The default colormaps are
defined in core.constants.

This texture has functionality for auto-resizing if it does not fit in
memory and padding if the system requires the size to be a factor of two.

"""

import OpenGL.GL as gl
import numpy as np

from visvis.core.misc import getOpenGlCapable, PropWithDraw, Range


# Dict that maps numpy datatypes to openGL data types
dtypes = {  'uint8':gl.GL_UNSIGNED_BYTE,    'int8':gl.GL_BYTE,
            'uint16':gl.GL_UNSIGNED_SHORT,  'int16':gl.GL_SHORT,
            'uint32':gl.GL_UNSIGNED_INT,    'int32':gl.GL_INT,
            'float32':gl.GL_FLOAT }


def makePowerOfTwo(data, ndim):
    """ makePowerOfTwo(data, ndim)
    
    If necessary, pad the data with zeros, to make the shape
    a power of two. If it already is shaped ok, the original data
    is returned.
    
    Use this function for systems with OpenGl < 2.0.
    
    """
    def nearestN(n1):
        n2 = 2
        while n2 < n1:
            n2*=2
        return n2
    
    # get old and new shape
    s1 = [n for n in data.shape]
    s2 = [nearestN(n) for n in data.shape]
    s2[ndim:] = s1[ndim:] # for color images
    
    # if not required return original
    if s1 == s2:
        return data
    
    # create empty image
    data2 = np.zeros(s2,dtype=data.dtype)
    
    # fill in the original data
    if ndim==1:
        data2[:s1[0]] = data
    elif ndim==2:
        data2[:s1[0],:s1[1]] = data
    elif ndim==3:
        data2[:s1[0],:s1[1],:s1[2]] = data
    else:
        raise ValueError("Cannot downsample data of this dimension.")
    return data2


def downSample(data, ndim):
    """ downSample(data, ndim)
    
    Downsample the data. Peforming a simple form of smoothing to prevent
    aliasing.
    
    """
    
    if ndim==1:
        # Decimate
        data2 = data[::2] * 0.4
        # Average in x
        tmp = data[1::2] * 0.3
        data2[:tmp.shape[0]] += tmp
        data2[1:] += tmp[:data2.shape[0]-1]
    elif ndim==2:
        # Decimate
        data2 = data[::2,::2] * 0.4
        # Average in y
        tmp = data[1::2,::2] * 0.15
        data2[:tmp.shape[0],:] += tmp
        data2[1:,:] += tmp[:data2.shape[0]-1,:]
        # Average in x
        tmp = data[::2,1::2] * 0.15
        data2[:,:tmp.shape[1]] += tmp
        data2[:,1:] += tmp[:,:data2.shape[1]-1]
    elif ndim==3:
        # Decimate
        data2 = data[::2,::2,::2] * 0.4
        # Average in z
        tmp = data[1::2,::2,::2] * 0.1
        data2[:tmp.shape[0],:,:] += tmp
        data2[1:,:,:] += tmp[:data2.shape[0]-1,:,:]
        # Average in y
        tmp = data[::2,1::2,::2] * 0.1
        data2[:,:tmp.shape[1],:] += tmp
        data2[:,1:,:] += tmp[:,:data2.shape[1]-1,:]
        # Average in x
        tmp = data[::2,::2,1::2] * 0.1
        data2[:,:,:tmp.shape[2]] += tmp
        data2[:,:,1:] += tmp[:,:,:data2.shape[2]-1]
    else:
        raise ValueError("Cannot downsample data of this dimension.")
    return data2



class TextureObject(object):
    """ TextureObject(texType)
    
    Basic texture class that wraps an OpenGl texture. It manages the OpenGl
    class and exposes a rather high-level interface to it.
    
    texType is one of gl.GL_TEXTURE_1D, gl.GL_TEXTURE_2D, gl.GL_TEXTURE_3D
    and specifies whether this is a 1D, 2D or 3D texture.
    
    Exposed methods:
      * Enable() call be for using
      * Disable() call after using
      * SetData() update the data
      * DestroyGl() remove only the texture from OpenGl memory.
      * Destroy() remove textures and reference to data.
        
    Note: this is not a Wobject nor a Wibject.
    
    """
    
    # One could argue to use polymorphism to implement 3 classes: one for
    # each dimension. Yes you could, but the way to handle the data and
    # communicate with OpenGl is so similar I chose not to. I use the
    # texType to determine which function to call.
    
    def __init__(self, ndim):
        
        # Check given texture type
        if ndim not in [1,2,3]:
            raise ValueError('Texture ndim should be 1, 2 or 3.')
        
        # Store the number of dimensions. This attribute is used to make the
        # choices for which OpenGl functions to use etc.
        self._ndim = ndim
        
        # Store the texture type, as we can determine it easily.
        tmp = {1:gl.GL_TEXTURE_1D, 2:gl.GL_TEXTURE_2D, 3:gl.GL_TEXTURE_3D}
        self._texType = tmp[ndim]
        
        # Texture ID. This is an integer by which OpenGl identifies the
        # texture.
        self._texId = 0
        
        # To store the used texture unit so we can disable it properly.
        self._texUnit = -1
        self._useTexUnit = False # set to True if OpenGl version high enough
        
        # A reference (not a weak one) to the original data as given with
        # SetData. We need this in order to re-upload the texture if it is
        # moved to another OpenGl context (other figure).
        # Note that the self._shape does not have to be self._dataRef.shape.
        self._dataRef = None
        
        # The shape of the data as uploaded to OpenGl. Is None if no
        # data was uploaded. Note that the self._shape does not have to
        # be self._dataRef.shape; the data might be downsampled.
        self._shape = None
        
        # A flag to indicate that the data in self._dataRef should be uploaded.
        # 1 signifies an update is required.
        # 2 signifies an update is required, with padding zeros.
        # -1 signifies the current data uploaded ok.
        # -2 ignifies the current data uploaded ok with padding.
        # 0 signifies failure of uploading
        self._uploadFlag = 1
        
        # Flag to indicate whether we can use this
        self._canUse = False
    
    
    def Enable(self, texUnit=0):
        """ Enable(texUnit)
        
        Enable the texture, using the given texture unit (max 9).
        If necessary, will upload/update the texture in OpenGl memory now.
        
        If texUnit is -1, will not bind the texture.
        
        """
        
        # Did we fail uploading texture last time?
        troubleLastTime = (self._uploadFlag==0)
        
        # If texture invalid, tell to upload, but only if we have a chance
        if self._texId == 0 or not gl.glIsTexture(self._texId):
            if not troubleLastTime:
                # Only if not in failure mode
                self._uploadFlag = abs(self._uploadFlag)
        
        # Store texture-Unit-id, and activate. Do before calling _setDataNow!
        if texUnit >= 0:
            self._texUnit = texUnit
            self._useTexUnit = getOpenGlCapable('1.3')
            if self._useTexUnit:
                gl.glActiveTexture( gl.GL_TEXTURE0 + texUnit )   # Opengl v1.3
        
        # If we should upload/update, do that now. (SetData also sets the flag)
        if self._uploadFlag > 0:
            self._SetDataNow()
        
        # check if ok now
        if not gl.glIsTexture(self._texId):
            if not troubleLastTime:
                print("Warning enabling texture, the texture is not valid. " +
                        "(Hiding message for future draws.)")
            return
        
        # Enable texturing, and bind to texture
        if texUnit >= 0:
            gl.glEnable(self._texType)
            gl.glBindTexture(self._texType, self._texId)
    
    
    def Disable(self):
        """ Disable()
        
        Disable the texture. It's safe to call this, even if the texture
        was not enabled.
        
        """
        
        # No need to disable. Also, if disabled because system does not
        # know 3D textures, we can not call glDisable with that arg.
        if self._uploadFlag == 0:
            return
        
        # Select active texture if we can
        if self._texUnit >= 0 and self._useTexUnit:
            gl.glActiveTexture( gl.GL_TEXTURE0 + self._texUnit )
            self._texUnit = -1
        
        # Disable
        gl.glDisable(self._texType)
        
        # Set active texture unit to default (0)
        if self._useTexUnit:
            gl.glActiveTexture( gl.GL_TEXTURE0 )
    
   
    def SetData(self, data):
        """ SetData(data)
        
        Set the data to display. If possible, will update the data in the
        existing texture (is possible if of the same shape).
        
        """
        
        # check data
        if not isinstance(data, np.ndarray):
            raise ValueError("Data should be a numpy array.")
        
        # check shape (raises ValueError if not ok)
        try:
            self._GetFormat(data.shape)
        except ValueError:
            raise # reraise from here
        
        # ok, store data and raise flag
        self._dataRef = data
        self._uploadFlag = abs(self._uploadFlag)
    
    
    def _SetDataNow(self):
        """ Make sure the data in self._dataRef is uploaded to
        OpenGl memory. If possible, update the data rather than
        create a new texture object.
        """
        
        # Test whether padding to a factor of two is required
        needPadding = (abs(self._uploadFlag) == 2)
        needPadding = needPadding or not getOpenGlCapable('2.0')
        
        # Set flag in case of failure (set to success at the end)
        # If we tried without padding, we can still try with padding.
        # Note: In theory, getOpenGlCapable('2.0') should be enough to
        # determine if padding is required. However, bloody ATI drivers
        # sometimes need 2**n textures even if OpenGl > 2.0. (I've
        # encountered this with someones PC and verified that the current
        # solution solves this.)
        if needPadding:
            self._uploadFlag = 0 # Give up
        else:
            self._uploadFlag = 2 # Try with padding next time
        
        # Get data.
        if self._dataRef is None:
            return
        data = self._dataRef
        
        # older OpenGl versions do not know about 3D textures
        if self._ndim==3 and not getOpenGlCapable('1.2','3D textures'):
            return
        
        # Convert data type to one supported by OpenGL
        if data.dtype.name not in dtypes:
            # Long integers become floats; int32 would not have enough range
            if data.dtype in (np.int64, np.uint64):
                data = data.astype(np.float32)
            # Bools become bytes
            elif data.dtype == np.bool:
                data = data.astype(np.uint8)
            else:
                # Make singles in all other cases (e.g. np.float64, np.float128)
                # We cannot explicitly use float128, since its not always defined
                data = data.astype(np.float32)
        
        # Determine type
        thetype = data.dtype.name
        if not thetype in dtypes:
            # this should not happen, since we convert incompatible types
            raise ValueError("Cannot convert datatype %s." % thetype)
        gltype = dtypes[thetype]
        
        # Determine format
        internalformat, format = self._GetFormat(data.shape)
        
        # Can we update or should we upload?
        
        if (    gl.glIsTexture(self._texId) and
                self._shape and (data.shape == self._shape) ):
            # We can update.
            
            # Bind to texture
            gl.glBindTexture(self._texType, self._texId)
            
            # update
            self._UpdateTexture(data, internalformat, format, gltype)
        
        else:
            # We should upload.
            
            # Remove any old data.
            self.DestroyGl()
            
            # Create texture object
            self._texId = gl.glGenTextures(1)
            
            # Bind to texture
            gl.glBindTexture(self._texType, self._texId)
            
            # Should we make the image a power of two?
            if needPadding:
                data2 = makePowerOfTwo(data, self._ndim)
                if data2 is not data:
                    data = data2
                    print("Warning: the data was padded to make it a power of two.")
            
            # test whether it fits, downsample if necessary
            ok, count = False, 0
            while not ok and count<8:
                ok = self._TestUpload(data, internalformat,format,gltype)
                if not ok:
                    #if count<2 and data.shape[0]<1000: # for testing
                    data = downSample(data, self._ndim)
                    count += 1
            
            # give warning or error
            if count and not ok:
                raise MemoryError(  "Could not upload texture to OpenGL, " +
                                    "even after 8 times downsampling.")
            elif count:
                print(  "Warning: data was downscaled " + str(count) +
                        " times to fit it in OpenGL memory." )
            
            # upload!
            self._UploadTexture(data, internalformat, format, gltype)
            
            # keep reference of data shape (as loaded to opengl)
            self._shape = data.shape
        
        # flag success
        if needPadding:
            self._uploadFlag = -2
        else:
            self._uploadFlag = -1
    
    
    def _UpdateTexture(self, data, internalformat, format, gltype):
        """ Update an existing texture object. It should have been
        checked whether this is possible (same shape).
        """
        
        # define dict
        D = {   1: (gl.glTexSubImage1D, gl.GL_TEXTURE_1D),
                2: (gl.glTexSubImage2D, gl.GL_TEXTURE_2D),
                3: (gl.glTexSubImage3D, gl.GL_TEXTURE_3D)}
        
        # determine function and target from texType
        uploadFun, target = D[self._ndim]
        
        # Build argument list
        shape = [i for i in reversed( list(data.shape[:self._ndim]) )]
        args = [target, 0] + [0 for i in shape] + shape + [format,gltype,data]
        
        # Upload!
        uploadFun(*tuple(args))
    
    
    def _TestUpload(self, data, internalformat, format, gltype):
        """ Test whether we can create a texture of the given shape.
        Returns True if we can, False if we can't.
        """
        
        # define dict
        D = {   1: (gl.glTexImage1D, gl.GL_PROXY_TEXTURE_1D),
                2: (gl.glTexImage2D, gl.GL_PROXY_TEXTURE_2D),
                3: (gl.glTexImage3D, gl.GL_PROXY_TEXTURE_3D)}
        
        # determine function and target from texType
        uploadFun, target = D[self._ndim]
        
        # build args list
        shape = [i for i in reversed( list(data.shape[:self._ndim]) )]
        args = [target, 0, internalformat] + shape + [0, format, gltype, None]
        
        # do fake upload
        uploadFun(*tuple(args))
        
        # test and return
        ok = gl.glGetTexLevelParameteriv(target, 0, gl.GL_TEXTURE_WIDTH)
        return bool(ok)
    
    
    def _UploadTexture(self, data, internalformat, format, gltype):
        """ Upload a texture to the current texture object.
        It should have been verified that the texture will fit.
        """
        
        # define dict
        D = {   1: (gl.glTexImage1D, gl.GL_TEXTURE_1D),
                2: (gl.glTexImage2D, gl.GL_TEXTURE_2D),
                3: (gl.glTexImage3D, gl.GL_TEXTURE_3D)}
        
        # determine function and target from texType
        uploadFun, target = D[self._ndim]
        
        # build args list
        shape = [i for i in reversed( list(data.shape[:self._ndim]) )]
        args = [target, 0, internalformat] + shape + [0, format, gltype, data]
        
        # call
        uploadFun(*tuple(args))
    
    
    def _GetFormat(self, shape):
        """ Get internalformat and format, based on the self._ndim
        and the shape. If the shape does not match with the texture
        type, an exception is raised.
        """
        
        if self._ndim == 1:
            if len(shape)==1:
                iformat, format = gl.GL_LUMINANCE8, gl.GL_LUMINANCE
            elif len(shape)==2 and shape[1] == 1:
                iformat, format = gl.GL_LUMINANCE8, gl.GL_LUMINANCE
            elif len(shape)==2 and shape[1] == 3:
                iformat, format = gl.GL_RGB, gl.GL_RGB
            elif len(shape)==2 and shape[1] == 4:
                iformat, format = gl.GL_RGBA, gl.GL_RGBA
            else:
                raise ValueError("Cannot create 1D texture, data of invalid shape.")
        
        elif self._ndim == 2:
        
            if len(shape)==2:
                iformat, format = gl.GL_LUMINANCE8, gl.GL_LUMINANCE
            elif len(shape)==3 and shape[2]==1:
                iformat, format = gl.GL_LUMINANCE8, gl.GL_LUMINANCE
            elif len(shape)==3 and shape[2]==3:
                iformat, format = gl.GL_RGB, gl.GL_RGB
            elif len(shape)==3 and shape[2]==4:
                iformat, format = gl.GL_RGBA, gl.GL_RGBA
            else:
                raise ValueError("Cannot create 2D texture, data of invalid shape.")
        
        elif self._ndim == 3:
        
            if len(shape)==3:
                iformat, format = gl.GL_LUMINANCE8, gl.GL_LUMINANCE
            elif len(shape)==4 and shape[3]==1:
                iformat, format = gl.GL_LUMINANCE8, gl.GL_LUMINANCE
            elif len(shape)==4 and shape[3]==3:
                iformat, format = gl.GL_RGB, gl.GL_RGB
            elif len(shape)==4 and shape[3]==4:
                iformat, format = gl.GL_RGBA, gl.GL_RGBA
            else:
                raise ValueError("Cannot create 3D texture, data of invalid shape.")
        
        else:
            raise ValueError("Cannot create a texture with these dimensions.")
        
        return iformat, format
    
    
    def DestroyGl(self):
        """ DestroyGl()
        
        Removes the texture from OpenGl memory. The internal reference
        to the original data is kept though.
        
        """
        try:
            if self._texId > 0:
                gl.glDeleteTextures([self._texId])
        except Exception:
            pass
        self._texId = 0
    
    
    def Destroy(self):
        """ Destroy()
        
        Really destroy data.
        
        """
        # remove OpenGl bits
        self.DestroyGl()
        # remove internal reference
        self._dataRef = None
        self._shape = None
    
    
    def __del__(self):
        self.Destroy()


class Colormap(TextureObject):
    """ Colormap()
    
    A colormap represents a table of colors to map
    grayscale data.
    
    """
    
    # Note that the OpenGL imaging subset also implements a colormap,
    # but it is not guaranteed that the subset is available.
    
    def __init__(self):
        TextureObject.__init__(self, 1)
        
        # CT: (0,0,0,0.0), (1,0,0,0.002), (0,0.5,1,0.6), (0,1,0,1)
        self._current = [(0,0,0), (1,1,1)]
        self.SetMap(self._current)
    
    
    def _UploadTexture(self, data, *args):
        """ Overloaded version to upload the texture.
        """
        
        # let the original class do the work
        TextureObject._UploadTexture(self, data, *args)
        
        # set interpolation and extrapolation parameters
        tmp = gl.GL_NEAREST # gl.GL_NEAREST | gl.GL_LINEAR
        gl.glTexParameteri(gl.GL_TEXTURE_1D, gl.GL_TEXTURE_MIN_FILTER, tmp)
        gl.glTexParameteri(gl.GL_TEXTURE_1D, gl.GL_TEXTURE_MAG_FILTER, tmp)
        gl.glTexParameteri(gl.GL_TEXTURE_1D, gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP)
    
    
    def GetMap(self):
        """ GetMap()
        
        Get the current texture map, as last set with SetMap().
        
        """
        return self._current
    
    
    def GetData(self):
        """ GetData()
        
        Get the full colormap as a 256x4 numpy array.
        
        """
        return self._dataRef
    
    
    def SetMap(self, *args):
        """ SetMap(*args)
        
        Set the colormap data. This method accepts several arguments:
        
        A list/tuple of tuples where each tuple represents a RGB or RGBA color.
        
        A dict with keys 'red', 'green', 'blue', 'alpha' (or only the first
        letter). Each dict should contain a list of 2-element tuples that
        specify index and color value. Indices should be between 0 and 1.
        
        A numpy array specifying the RGB or RGBA tuples.
        
        """
        
        # one argument given?
        if len(args)==1:
            args = args[0]
        
        # store
        self._current = args
        
        # init
        data = None
        
        # parse input
        
        if isinstance(args, dict):
            # DICT
            
            # Allow several color names
            for key in list(args.keys()):
                if key.lower() in ['r', 'red']:
                    args['r'] = args[key]
                elif key.lower() in ['g', 'green']:
                    args['g'] = args[key]
                if key.lower() in ['b', 'blue']:
                    args['b'] = args[key]
                if key.lower() in ['a', 'alpha']:
                    args['a'] = args[key]
            # Init data, alpha 1
            data2 = np.zeros((256,4),np.float32)
            data2[:,3] = 1.0
            # For each channel ...
            for i in range(4):
                channel = 'rgba'[i]
                if not channel in args:
                    continue
                # Get value list and check
                values = args[channel]
                if not hasattr(values,'__len__'):
                    raise ValueError('Invalid colormap.')
                # Init interpolation
                data = np.zeros((len(values),), dtype=np.float32)
                x = np.linspace(0.0, 1.0, 256)
                xp = np.zeros((len(values),), dtype=np.float32)
                # Insert values
                count = -1
                for el in values:
                    count += 1
                    if not hasattr(el,'__len__') or len(el) != 2:
                        raise ValueError('Colormap dict entries must have 2 elements.')
                    xp[count] = el[0]
                    data[count] = el[1]
                # Interpolate
                data2[:,i] = np.interp(x, xp, data)
            # Set
            data = data2
        
        elif isinstance(args, (tuple, list)):
            # LIST
            
            data = np.zeros((len(args),4), dtype=np.float32)
            data[:,3] = 1.0 # init alpha to be all ones
            count = -1
            for el in args:
                count += 1
                if not hasattr(el,'__len__') or len(el) not in [3,4]:
                    raise ValueError('Colormap entries must have 3 or 4 elements.')
                elif len(el)==3:
                    data[count,:] = el[0], el[1], el[2], 1.0
                elif len(el)==4:
                    data[count,:] = el[0], el[1], el[2], el[3]
        
        elif isinstance(args, np.ndarray):
            # ARRAY
            
            if args.ndim != 2 or args.shape[1] not in [3,4]:
                raise ValueError('Colormap entries must have 3 or 4 elements.')
            elif args.shape[1]==3:
                data = np.ones((args.shape[0], 4), dtype=np.float32)
                data[:, 0:3] = args
            elif args.shape[1]==4:
                data = args
            else:
                raise ValueError("Invalid argument to set colormap.")
        
        # Apply interpolation (if required)
        if data is not None:
            if data.shape[0] == 256 and data.dtype == np.float32:
                data2 = data
            else:
                # interpolate first
                x = np.linspace(0.0, 1.0, 256)
                xp = np.linspace(0.0, 1.0, data.shape[0])
                data2 = np.zeros((256,4),np.float32)
                for i in range(4):
                    data2[:,i] = np.interp(x, xp, data[:,i])
            # store texture
            #self._data = data2
            self.SetData(data2)


class Colormapable(object):
    """ Mixer class for wobjects that can map scalar data to colors.
    
    Instances of this class have a colormap propery and a clim property.
    
    Inheriting classes can implement the following methods to overload
    the default behavior:
      * _GetColormap, _SetColormap(cmap)
      * _GetClim, _SetClim(clim)
      * _EnableColormap(texUnit=0), _DisableColormap
    
    """
    
    def __init__(self):
        self._colormap = Colormap()
        self._clim = Range(0,1)
    
    @PropWithDraw
    def colormap():
        """ Get/get the colormap. This can be:
          * A tuple/list of element which each have 3 or 4 values (RGB/RGBA)
          * A Nx3 or Nx4 numpy array
          * A dict with names R,G,B,A, where each value is a list. Each
            element in the list is a tuple (index, color-value).
        
        Visvis defines a number of standard colormaps named by vv.CM_*. A few
        examples are CM_GRAY, CM_HOT, CM_JET, CM_SUMMER, CM_CT1. See
        vv.colormaps for a dict with all available default colormaps.
        """
        def fget(self):
            return self._GetColormap()
        def fset(self, value):
            self._SetColormap(value)
        return locals()
    
    @PropWithDraw
    def clim():
        """ Get/set the contrast limits. For a gray colormap, clim.min
        is black, clim.max is white.
        """
        def fget(self):
            return self._GetClim()
        def fset(self, value):
            if not isinstance(value, Range):
                value = Range(value)
            self._SetClim(value)
        return locals()
    
    def _GetColormap(self):
        return self._colormap.GetMap()
    def _SetColormap(self, value):
        self._colormap.SetMap(value)
    
    def _EnableColormap(self, texUnit=0):
        self._colormap.Enable(texUnit)
    def _DisableColormap(self):
        self._colormap.Disable()
    
    def _GetClim(self):
        return self._clim
    def _SetClim(self, value):
        self._clim = value



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import visvis as vv
import numpy as np
from visvis.utils.pypoints import Point, Pointset


# Alternative implementation bu subdeivinding a tetrahedon,
# but not as good for mapping textures.
def getSphere(ndiv=3, radius=1.0):
    # Example taken from the Red book, end of chaper 2.
    
    # Define constants
    X = 0.525731112119133606
    Z = 0.850650808352039932
    
    # Creta vdata
    vdata = Pointset(3)
    app = vdata.append
    app(-X, 0.0, Z); app(X, 0.0, Z); app(-X, 0.0, -Z); app(X, 0.0, -Z)
    app(0.0, Z, X); app(0.0, Z, -X); app(0.0, -Z, X); app(0.0, -Z, -X)
    app(Z, X, 0.0); app(-Z, X, 0.0); app(Z, -X, 0.0); app(-Z, -X, 0.0)
    
    # Create faces
    tindices = [
        [0,4,1], [0,9,4], [9,5,4], [4,5,8], [4,8,1],
        [8,10,1], [8,3,10], [5,3,8], [5,2,3], [2,7,3],
        [7,10,3], [7,6,10], [7,11,6], [11,0,6], [0,1,6],
        [6,1,10], [9,0,11], [9,11,2], [9,2,5], [7,2,11] ]
    tindices = np.array(tindices, dtype=np.uint32)
    
    # Init vertex array
    vertices = Pointset(3)
    
    # Define function to recursively create vertices and normals
    def drawtri(a, b, c, div):
        if (div<=0):
            vertices.append(a)
            vertices.append(b)
            vertices.append(c)
        else:
            ab = Point(0,0,0)
            ac = Point(0,0,0)
            bc = Point(0,0,0)
            for i in range(3):
                ab[i]=(a[i]+b[i])/2.0
                ac[i]=(a[i]+c[i])/2.0
                bc[i]=(b[i]+c[i])/2.0
            ab = ab.normalize(); ac = ac.normalize(); bc = bc.normalize()
            drawtri(a, ab, ac, div-1)
            drawtri(b, bc, ab, div-1)
            drawtri(c, ac, bc, div-1)
            drawtri(ab, bc, ac, div-1)
    
    # Create vertices
    for i in range(20):
        drawtri(    vdata[int(tindices[i][0])],
                    vdata[int(tindices[i][1])],
                    vdata[int(tindices[i][2])],
                    ndiv )
    
    # Create normals and scale vertices
    normals = vertices.copy()
    vertices *= radius
    
    # Done
    return vertices, normals


# Alternative implementation bu subdeivinding a tetrahedon,
# but not as good for mapping textures.
# This version uses faces to reuse vertices.
def getSphereWithFaces(ndiv=3, radius=1.0):
    # Example taken from the Red book, end of chaper 2.
    
    # Define constants
    X = 0.525731112119133606
    Z = 0.850650808352039932
    
    # Creta vdata
    vdata = Pointset(3)
    app = vdata.append
    app(-X, 0.0, Z); app(X, 0.0, Z); app(-X, 0.0, -Z); app(X, 0.0, -Z)
    app(0.0, Z, X); app(0.0, Z, -X); app(0.0, -Z, X); app(0.0, -Z, -X)
    app(Z, X, 0.0); app(-Z, X, 0.0); app(Z, -X, 0.0); app(-Z, -X, 0.0)
    
    # Create faces
    tindices = [
        [0,4,1], [0,9,4], [9,5,4], [4,5,8], [4,8,1],
        [8,10,1], [8,3,10], [5,3,8], [5,2,3], [2,7,3],
        [7,10,3], [7,6,10], [7,11,6], [11,0,6], [0,1,6],
        [6,1,10], [9,0,11], [9,11,2], [9,2,5], [7,2,11] ]
    tindices = np.array(tindices, dtype=np.uint32)
    
    # Init vertex array with existing points, init faces as empty list
    vertices = vdata.copy()
    faces = []
    
    # Define function to recursively create vertices and normals
    def drawtri(ia, ib, ic, div):
        a, b, c = vertices[ia] , vertices[ib], vertices[ic]
        if (div<=0):
            # Store faces here
            faces.extend([ia, ib, ic])
        else:
            # Create new points
            ab = Point(0,0,0)
            ac = Point(0,0,0)
            bc = Point(0,0,0)
            for i in range(3):
                ab[i]=(a[i]+b[i])/2.0
                ac[i]=(a[i]+c[i])/2.0
                bc[i]=(b[i]+c[i])/2.0
            ab = ab.normalize(); ac = ac.normalize(); bc = bc.normalize()
            # Add new points
            i_offset = len(vertices)
            vertices.append(ab)
            vertices.append(ac)
            vertices.append(bc)
            iab, iac, ibc = i_offset+0, i_offset+1, i_offset+2
            #
            drawtri(ia, iab, iac, div-1)
            drawtri(ib, ibc, iab, div-1)
            drawtri(ic, iac, ibc, div-1)
            drawtri(iab, ibc, iac, div-1)
    
    # Create vertices
    for i in range(20):
        drawtri(    int(tindices[i][0]),
                    int(tindices[i][1]),
                    int(tindices[i][2]),
                    ndiv )
    
    # Create normals and scale vertices
    normals = vertices.copy()
    vertices *= radius
    
    # Create faces
    faces = np.array(faces, dtype='uint32')
    
    # Done
    return vertices, faces, normals


def solidSphere(translation=None, scaling=None, direction=None, rotation=None,
                N=16, M=16, axesAdjust=True, axes=None):
    """ solidSphere(translation=None, scaling=None, direction=None, rotation=None,
                    N=16, M=16, axesAdjust=True, axes=None)
    
    Creates a solid sphere with quad faces and centered at the origin.
    Returns an OrientableMesh instance.
    
    Parameters
    ----------
    Note that translation, scaling, and direction can also be given
    using a Point instance.
    translation : (dx, dy, dz), optional
        The translation in world units of the created world object.
    scaling: (sx, sy, sz), optional
        The scaling in world units of the created world object.
    direction: (nx, ny, nz), optional
        Normal vector that indicates the direction of the created world object.
    rotation: scalar, optional
        The anle (in degrees) to rotate the created world object around its
        direction vector.
    N : int
        The number of subdivisions around the Z axis (similar to lines
        of longitude). If smaller than 8, flat shading is used instead
        of smooth shading.
    M : int
        The number of subdivisions along the Z axis (similar to lines
        of latitude). If smaller than 8, flat shading is used instead
        of smooth shading.
    axesAdjust : bool
        If True, this function will call axes.SetLimits(), and set
        the camera type to 3D. If daspectAuto has not been set yet,
        it is set to False.
    axes : Axes instance
        Display the bars in the given axes, or the current axes if not given.
    
    """
    
    # Note that the number of vertices around the axis is N+1. This
    # would not be necessary per see, but it helps create a nice closed
    # texture when it is mapped. There are N number of faces though.
    # Similarly, to obtain M faces along the axis, we need M+1
    # vertices.
    
    # Quick access
    pi2 = np.pi*2
    cos = np.cos
    sin = np.sin
    sl = N+1
    
    # Calculate vertices, normals and texcords
    vertices = Pointset(3)
    normals = Pointset(3)
    texcords = Pointset(2)
    # Cone
    for m in range(M+1):
        a = np.pi * float(m) / M
        z = cos(a)
        v = float(m)/M
        #
        for n in range(N+1):
            b = pi2 * float(n) / N
            u = float(n) / (N)
            x = cos(b) * sin(a)
            y = sin(b) * sin(a)
            vertices.append(x,y,z)
            normals.append(x,y,z)
            texcords.append(u,v)
    
    # Calculate indices
    indices = []
    for j in range(M):
        for i in range(N):
            #indices.extend([j*sl+i, j*sl+i+1, (j+1)*sl+i+1, (j+1)*sl+i])
            indices.extend([(j+1)*sl+i, (j+1)*sl+i+1, j*sl+i+1, j*sl+i])
    
    # Make indices a numpy array
    indices = np.array(indices, dtype=np.uint32)
    
    
    ## Visualize
    
    # Create axes
    if axes is None:
        axes = vv.gca()
    
    # Create mesh
    m = vv.OrientableMesh(axes, vertices, indices, normals, texcords, 4)
    #
    if translation is not None:
        m.translation = translation
    if scaling is not None:
        m.scaling = scaling
    if direction is not None:
        m.direction = direction
    if rotation is not None:
        m.rotation = rotation
    
    # Adjust axes
    if axesAdjust:
        if axes.daspectAuto is None:
            axes.daspectAuto = False
        axes.cameraType = '3d'
        axes.SetLimits()
    
    # Done
    axes.Draw()
    return m


if __name__ == '__main__':
    m = vv.solidSphere(scaling=(1,1,1.5), direction=(1,1,3))
    m.SetTexture( vv.imread('astronaut.png') )



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Robert Schroll
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" The GTK backend.
"""

import os
import sys

import visvis
from visvis import BaseFigure, events, constants
from visvis.core.misc import getResourceDir

import gtk
import gtk.gtkgl
import gobject

import OpenGL.GL as gl

MOUSEMAP = {gtk.gdk.BUTTON_PRESS   : 'down',
            gtk.gdk.BUTTON_RELEASE : 'up',
            gtk.gdk._2BUTTON_PRESS : 'double'}

KEYMAP = {  gtk.keysyms.Shift_L: constants.KEY_SHIFT,
            gtk.keysyms.Shift_R: constants.KEY_SHIFT,
            gtk.keysyms.Alt_L: constants.KEY_ALT,
            gtk.keysyms.Alt_R: constants.KEY_ALT,
            gtk.keysyms.Control_L: constants.KEY_CONTROL,
            gtk.keysyms.Control_R: constants.KEY_CONTROL,
            gtk.keysyms.Left: constants.KEY_LEFT,
            gtk.keysyms.Up: constants.KEY_UP,
            gtk.keysyms.Right: constants.KEY_RIGHT,
            gtk.keysyms.Down: constants.KEY_DOWN,
            gtk.keysyms.Page_Up: constants.KEY_PAGEUP,
            gtk.keysyms.Page_Down: constants.KEY_PAGEDOWN,
            gtk.keysyms.KP_Enter: constants.KEY_ENTER,
            gtk.keysyms.Return: constants.KEY_ENTER,
            gtk.keysyms.Escape: constants.KEY_ESCAPE,
            gtk.keysyms.Delete: constants.KEY_DELETE,
            }

# Make uppercase letters be lowercase
for i in range(ord('A'), ord('Z')):
    KEYMAP[i] = i+32


def modifiers(event):
    """Convert the GTK state into a tuple of active modifier keys."""
    if not hasattr(event, 'state'):
        return ()
    
    mods = ()
    if event.state & gtk.gdk.SHIFT_MASK:
        mods += constants.KEY_SHIFT,
    if event.state & gtk.gdk.CONTROL_MASK:
        mods += constants.KEY_CONTROL,
    if event.state & gtk.gdk.MOD1_MASK:
        mods += constants.KEY_ALT,
    return mods

class GlCanvas(gtk.gtkgl.DrawingArea):
    
    def __init__(self, figure, *args, **kw):
        gtk.gtkgl.DrawingArea.__init__(self)
        self.add_events(gtk.gdk.BUTTON_PRESS_MASK |
                        gtk.gdk.BUTTON_RELEASE_MASK |
                        gtk.gdk.POINTER_MOTION_MASK |
                        gtk.gdk.POINTER_MOTION_HINT_MASK |
                        gtk.gdk.KEY_PRESS_MASK |
                        gtk.gdk.KEY_RELEASE_MASK |
                        gtk.gdk.ENTER_NOTIFY_MASK |
                        gtk.gdk.LEAVE_NOTIFY_MASK |
                        gtk.gdk.FOCUS_CHANGE_MASK)
        self.set_property('can-focus', True)
        
        self.figure = figure
        
        # Configure OpenGL framebuffer.
        # Try to get a double-buffered framebuffer configuration,
        # if not successful then try to get a single-buffered one.
        display_mode = (gtk.gdkgl.MODE_RGB    |
                        gtk.gdkgl.MODE_DEPTH  |
                        gtk.gdkgl.MODE_DOUBLE)
        try:
            glconfig = gtk.gdkgl.Config(mode=display_mode)
        except gtk.gdkgl.NoMatches:
            display_mode &= ~gtk.gdkgl.MODE_DOUBLE
            glconfig = gtk.gdkgl.Config(mode=display_mode)
        self.set_gl_capability(glconfig)
        
        # Connect the relevant signals.
        self.connect('configure_event', self._on_configure_event)
        self.connect('expose_event',    self._on_expose_event)
        self.connect('delete_event', self._on_delete_event)
        self.connect('motion_notify_event', self._on_motion_notify_event)
        self.connect('button_press_event', self._on_button_event)
        self.connect('button_release_event', self._on_button_event)
        self.connect('scroll_event', self._on_scroll_event)
        self.connect('key_press_event', self._on_key_press_event)
        self.connect('key_release_event', self._on_key_release_event)
        self.connect('enter_notify_event', self._on_enter_notify_event)
        self.connect('leave_notify_event', self._on_leave_notify_event)
        self.connect('focus_in_event', self._on_focus_in_event)
        
    def _on_configure_event(self, *args):
        if self.figure:
            self.figure._OnResize()
        return False
        
    def _on_delete_event(self, *args):
        if self.figure:
            self.figure.Destroy()
        return True # Let figure.Destoy() destroy this widget
    
    def _on_motion_notify_event(self, widget, event):
        if event.is_hint:
            x, y, state = event.window.get_pointer()  # noqa
        else:
            x, y, state = event.x, event.y, event.state  # noqa
        if self.figure:
            self.figure._GenerateMouseEvent('motion', x, y, 0, modifiers(event))
    
    def _on_button_event(self, widget, event):
        if event.type in MOUSEMAP:
            button = {1:1, 3:2}.get(event.button, 0)
            self.figure._GenerateMouseEvent(MOUSEMAP[event.type], event.x, event.y,
                                            button, modifiers(event))
    
    def _on_scroll_event(self, widget, event):
        horizontal = {gtk.gdk.SCROLL_LEFT: 1.0, gtk.gdk.SCROLL_RIGHT: -1.0}.get(event.direction, 0)
        vertical = {gtk.gdk.SCROLL_UP: 1.0, gtk.gdk.SCROLL_DOWN: -1.0}.get(event.direction, 0)
        if horizontal or vertical:
            self.figure._GenerateMouseEvent('scroll', event.x, event.y,
                                            horizontal, vertical, modifiers(event))
    
    def _on_key_press_event(self, widget, event):
        self.figure._GenerateKeyEvent('keydown', KEYMAP.get(event.keyval, event.keyval),
                                      event.string, modifiers(event))
    
    def _on_key_release_event(self, widget, event):
        self.figure._GenerateKeyEvent('keyup', KEYMAP.get(event.keyval, event.keyval),
                                      event.string, modifiers(event))
    
    def _on_enter_notify_event(self, widget, event):
        if self.figure:
            ev = self.figure.eventEnter
            ev.Set(0,0,0)
            ev.Fire()
    
    def _on_leave_notify_event(self, widget, event):
        if self.figure:
            ev = self.figure.eventLeave
            ev.Set(0,0,0)
            ev.Fire()
    
    def _on_focus_in_event(self, widget, event):
        if self.figure:
            BaseFigure._currentNr = self.figure.nr
   
    def _on_expose_event(self, *args):
        # Obtain a reference to the OpenGL drawable
        # and rendering context.
        gldrawable = self.get_gl_drawable()
        glcontext = self.get_gl_context()

        # OpenGL begin
        if not gldrawable.gl_begin(glcontext):
            return False

        self.figure.OnDraw()

        # OpenGL end
        gldrawable.gl_end()
    
    def set_current(self):
        gldrawable = self.get_gl_drawable()
        glcontext = self.get_gl_context()
        
        gldrawable.make_current(glcontext)
    
    def swap_buffers(self):
        gldrawable = self.get_gl_drawable()
        glcontext = self.get_gl_context()  # noqa

        if gldrawable.is_double_buffered():
            gldrawable.swap_buffers()
        else:
            gl.glFlush()


class Figure(BaseFigure):
    
    def __init__(self, *args, **kw):
        self._widget = None
        self._widget_args = (args, kw)
        if kw.get('create_widget', True):
            self.CreateWidget()
        
        BaseFigure.__init__(self)
    
    def CreateWidget(self):
        """Create the Figure's widget if necessary, and return the
        widget."""
        if self._widget is None:
            # Make sure there is a native app and the timer is started
            # (also when embedded)
            app.Create()
            
            # create gl widget
            updatePosition = False
            args, kwargs = self._widget_args
            if 'create_widget' in kwargs:
                updatePosition = True
                del(kwargs['create_widget'])
            self._widget = GlCanvas(self, *args, **kwargs)
            if updatePosition:
                self.position._Changed()
        return self._widget
    
    def _SetCurrent(self):
        """Make figure the current OpenGL context."""
        if self._widget and not self._destroyed:
            self._widget.set_current()
    
    def _SwapBuffers(self):
        """Swap the memory and screen buffers."""
        if self._widget and not self._destroyed:
            self._widget.swap_buffers()
    
    def _RedrawGui(self):
        """Make the widget redraw itself."""
        if self._widget:
            self._widget.queue_draw()
    
    def _ProcessGuiEvents(self):
        """Process all events in queue."""
        app.ProcessEvents()
    
    def _SetTitle(self, title):
        """Set the title, when not used in application."""
        if self._widget and not self._destroyed:
            window = self._widget.parent
            if isinstance(window, gtk.Window):
                window.set_title(title)
    
    def _SetPosition(self, x, y, w, h):
        """Set the position and size of the widget.  If it is embedded,
        ignore the x and y coordinates."""
        if self._widget and not self._destroyed:
            self._widget.set_size_request(w, h)
            self._widget.queue_resize()
            window = self._widget.parent
            if isinstance(window, gtk.Window):
                window.move(x, y)
                window.resize(w, h)
    
    def _GetPosition(self):
        """Get the widget's position."""
        if self._widget and not self._destroyed:
            alloc = self._widget.allocation
            x, y = alloc.x, alloc.y
            window = self._widget.parent
            if isinstance(window, gtk.Window):
                x, y = window.get_position()
            return x, y, alloc.width, alloc.height
        return 0, 0, 0, 0
    
    def _Close(self, widget):
        """Close the widget."""
        if widget is None:
            widget = self._widget
        if widget:
            window = widget.parent
            # The destroy() method causes IPython to emit on error on my system
            # the first time it happens (almar)
            if isinstance(window, gtk.Window):
                window.destroy()
            else:
                widget.destroy()
        
        # If no more figures, quit
        # If in script-mode, we nicely quit. If in interactive mode, we won't.
        if len(BaseFigure._figures) == 0:
            app.Quit()


def newFigure():
    """Create a figure and put it in a window."""
    
    figure = Figure()
    window = gtk.Window()
    
    # Set icon
    try:
        iconfile = os.path.join(getResourceDir(), 'visvis_icon_gtk.png')
        window.set_icon_from_file(iconfile)
    except Exception:
        pass
    
    # From GTKGL example
    if sys.platform != 'win32':
        window.set_resize_mode(gtk.RESIZE_IMMEDIATE)
    window.set_reallocate_redraws(True)
    
    window.add(figure._widget)
    size = visvis.settings.figureSize
    figure._widget.set_size_request(size[0], size[1])
    window.set_geometry_hints(min_width=100, min_height=100)
    window.show_all()
    
    window.connect('delete-event', figure._widget._on_delete_event)
    
    # Initialize OpenGl
    figure.DrawNow()
    return figure


class VisvisEventsTimer:
    """ Timer that can be started and stopped.
    """
    
    def __init__(self):
        self._running = False
    
    def Start(self):
        if not self._running:
            self._running = True
            self._PostTimeout()
    
    def Stop(self):
        self._running = False
    
    def _PostTimeout(self):
        gobject.timeout_add(10, self._Fire)
    
    def _Fire(self):
        if self._running:
            events.processVisvisEvents()
            return True # So called again.
        else:
            return False # Stop timer.


class App(events.App):
    """App()
    
    Application class to wrap the GUI applications in a class
    with a simple interface that is the same for all backends.
    
    This is the GTK implementation.
    
    """
    
    def __init__(self):
        # Create timer
        self._timer = VisvisEventsTimer()
    
    def _GetNativeApp(self):
        """Ensure the GTK app exists."""
        
        # Start timer
        self._timer.Start()
        
        # Prevent quiting when used interactively
        if not hasattr(gtk, 'vv_do_quit'):
            gtk.vv_do_quit = False
        
        # Return singleton gtk object, which represents the gtk application
        return gtk
    
    def _ProcessEvents(self):
        """Process GTK events."""
        gtk.gdk.threads_enter() # enter/leave prevents IPython -gthread to hang
        while gtk.events_pending():
            gtk.main_iteration(False)
        gtk.gdk.threads_leave()
    
    def _Run(self):
        """Enter GTK mainloop."""
        self._GetNativeApp()
        
        if gtk.main_level() == 0:
            # We need to start the mainloop.  This means we will also
            # have to kill the mainloop when the last figure is closed.
            gtk.vv_do_quit = True
            gtk.main()
    
    def Quit(self):
        if gtk.vv_do_quit: # We started the mainloop, so we better kill it.
            gtk.main_quit()

app = App()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# flake8: noqa
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
Freetype enum types
-------------------

FT_PIXEL_MODES: An enumeration type used to describe the format of pixels in a
                given bitmap. Note that additional formats may be added in the
                future.

FT_GLYPH_BBOX_MODES: The mode how the values of FT_Glyph_Get_CBox are returned.

FT_GLYPH_FORMATS: An enumeration type used to describe the format of a given
                  glyph image. Note that this version of FreeType only supports
                  two image formats, even though future font drivers will be
                  able to register their own format.

FT_ENCODINGS: An enumeration used to specify character sets supported by
              charmaps. Used in the FT_Select_Charmap API function.

FT_RENDER_MODES: An enumeration type that lists the render modes supported by
                 FreeType 2. Each mode corresponds to a specific type of
                 scanline conversion performed on the outline.

FT_LOAD_TARGETS: A list of values that are used to select a specific hinting
                 algorithm to use by the hinter. You should OR one of these
                 values to your 'load_flags' when calling FT_Load_Glyph.

FT_LOAD_FLAGS: A list of bit-field constants used with FT_Load_Glyph to
               indicate what kind of operations to perform during glyph
               loading.

FT_STYLE_FLAGS: A list of bit-flags used to indicate the style of a given
                face. These are used in the 'style_flags' field of FT_FaceRec.

FT_FACE_FLAGS: A list of bit flags used in the 'face_flags' field of the
               FT_FaceRec structure. They inform client applications of
               properties of the corresponding face.

FT_OUTLINE_FLAGS: A list of bit-field constants use for the flags in an
                  outline's 'flags' field.

FT_OPEN_MODES: A list of bit-field constants used within the 'flags' field of
               the FT_Open_Args structure.

FT_KERNING_MODES: An enumeration used to specify which kerning values to return
                  in FT_Get_Kerning.

FT_STROKER_LINEJOINS: These values determine how two joining lines are rendered
                      in a stroker.

FT_STROKER_LINECAPS: These values determine how the end of opened sub-paths are
                     rendered in a stroke.

FT_STROKER_BORDERS: These values are used to select a given stroke border in
                    FT_Stroker_GetBorderCounts and FT_Stroker_ExportBorder.

FT_LCD_FILTERS: A list of values to identify various types of LCD filters.

TT_PLATFORMS: A list of valid values for the 'platform_id' identifier code in
              FT_CharMapRec and FT_SfntName structures.

TT_APPLE_IDS: A list of valid values for the 'encoding_id' for
              TT_PLATFORM_APPLE_UNICODE charmaps and name entries.
 
TT_MAC_IDS: A list of valid values for the 'encoding_id' for
            TT_PLATFORM_MACINTOSH charmaps and name entries.

TT_MS_IDS: A list of valid values for the 'encoding_id' for
           TT_PLATFORM_MICROSOFT charmaps and name entries.

TT_ADOBE_IDS: A list of valid values for the 'encoding_id' for
              TT_PLATFORM_ADOBE charmaps. This is a FreeType-specific
              extension!

TT_MAC_LANGIDS: Possible values of the language identifier field in the name
                records of the TTF `name' table if the `platform' identifier
                code is TT_PLATFORM_MACINTOSH.

TT_MS_LANGIDS: Possible values of the language identifier field in the name
               records of the TTF `name' table if the `platform' identifier
               code is TT_PLATFORM_MICROSOFT.

TT_NAME_IDS: Possible values of the `name' identifier field in the name
             records of the TTF `name' table.  These values are platform
             independent.
'''

# -----------------------------------------------------------------------------
# An enumeration type used to describe the format of pixels in a given
# bitmap. Note that additional formats may be added in the future.
#
# FT_PIXEL_MODE_NONE
# Value 0 is reserved.
#
# FT_PIXEL_MODE_MONO
# A monochrome bitmap, using 1 bit per pixel. Note that pixels are stored in
# most-significant order (MSB), which means that the left-most pixel in a byte
# has value 128.
#
# FT_PIXEL_MODE_GRAY
# An 8-bit bitmap, generally used to represent anti-aliased glyph images. Each
# pixel is stored in one byte. Note that the number of 'gray' levels is stored
# in the 'num_grays' field of the FT_Bitmap structure (it generally is 256).
#
# FT_PIXEL_MODE_GRAY2
# A 2-bit per pixel bitmap, used to represent embedded anti-aliased bitmaps in
# font files according to the OpenType specification. We haven't found a single
# font using this format, however.
#
# FT_PIXEL_MODE_GRAY4
# A 4-bit per pixel bitmap, representing embedded anti-aliased bitmaps in font
# files according to the OpenType specification. We haven't found a single font
# using this format, however.
#
# FT_PIXEL_MODE_LCD
# An 8-bit bitmap, representing RGB or BGR decimated glyph images used for
# display on LCD displays; the bitmap is three times wider than the original
# glyph image. See also FT_RENDER_MODE_LCD.
#
# FT_PIXEL_MODE_LCD_V
# An 8-bit bitmap, representing RGB or BGR decimated glyph images used for
# display on rotated LCD displays; the bitmap is three times taller than the
# original glyph image. See also FT_RENDER_MODE_LCD_V.
#
FT_PIXEL_MODES = {'FT_PIXEL_MODE_NONE' : 0,
                  'FT_PIXEL_MODE_MONO' : 1,
                  'FT_PIXEL_MODE_GRAY' : 2,
                  'FT_PIXEL_MODE_GRAY2': 3,
                  'FT_PIXEL_MODE_GRAY4': 4,
                  'FT_PIXEL_MODE_LCD'  : 5,
                  'FT_PIXEL_MODE_LCD_V': 6,
                  'FT_PIXEL_MODE_MAX'  : 7}
globals().update(FT_PIXEL_MODES)
ft_pixel_mode_none  = FT_PIXEL_MODE_NONE
ft_pixel_mode_mono  = FT_PIXEL_MODE_MONO
ft_pixel_mode_grays = FT_PIXEL_MODE_GRAY
ft_pixel_mode_pal2  = FT_PIXEL_MODE_GRAY2
ft_pixel_mode_pal4  = FT_PIXEL_MODE_GRAY4


# -----------------------------------------------------------------------------
# The mode how the values of FT_Glyph_Get_CBox are returned.
#
# FT_GLYPH_BBOX_UNSCALED
# Return unscaled font units.
#
# FT_GLYPH_BBOX_SUBPIXELS
# Return unfitted 26.6 coordinates.
#
# FT_GLYPH_BBOX_GRIDFIT
# Return grid-fitted 26.6 coordinates.
#
# FT_GLYPH_BBOX_TRUNCATE
# Return coordinates in integer pixels.
#
# FT_GLYPH_BBOX_PIXELS
# Return grid-fitted pixel coordinates.
#
FT_GLYPH_BBOX_MODES = {'FT_GLYPH_BBOX_UNSCALED'  : 0,
                       'FT_GLYPH_BBOX_SUBPIXELS' : 0,
                       'FT_GLYPH_BBOX_GRIDFIT'   : 1,
                       'FT_GLYPH_BBOX_TRUNCATE'  : 2,
                       'FT_GLYPH_BBOX_PIXELS'    : 3}
globals().update(FT_GLYPH_BBOX_MODES)



# -----------------------------------------------------------------------------
# An enumeration type used to describe the format of a given glyph image. Note
# that this version of FreeType only supports two image formats, even though
# future font drivers will be able to register their own format.
#
# FT_GLYPH_FORMAT_NONE
# The value 0 is reserved.

# FT_GLYPH_FORMAT_COMPOSITE
# The glyph image is a composite of several other images. This format is only
# used with FT_LOAD_NO_RECURSE, and is used to report compound glyphs (like
# accented characters).
#
# FT_GLYPH_FORMAT_BITMAP
# The glyph image is a bitmap, and can be described as an FT_Bitmap. You
# generally need to access the 'bitmap' field of the FT_GlyphSlotRec structure
# to read it.
#
# FT_GLYPH_FORMAT_OUTLINE
# The glyph image is a vectorial outline made of line segments and Bezier arcs;
# it can be described as an FT_Outline; you generally want to access the
# 'outline' field of the FT_GlyphSlotRec structure to read it.
#
# FT_GLYPH_FORMAT_PLOTTER
# The glyph image is a vectorial path with no inside and outside contours. Some
# Type 1 fonts, like those in the Hershey family, contain glyphs in this
# format. These are described as FT_Outline, but FreeType isn't currently
# capable of rendering them correctly.
#
def _FT_IMAGE_TAG(a,b,c,d):
    return ( ord(a) << 24 | ord(b) << 16 | ord(c) << 8 | ord(d) )
FT_GLYPH_FORMATS = {
    'FT_GLYPH_FORMAT_NONE'      : _FT_IMAGE_TAG( '\0','\0','\0','\0' ),
    'FT_GLYPH_FORMAT_COMPOSITE' : _FT_IMAGE_TAG( 'c','o','m','p' ),
    'FT_GLYPH_FORMAT_BITMAP'    : _FT_IMAGE_TAG( 'b','i','t','s' ),
    'FT_GLYPH_FORMAT_OUTLINE'   : _FT_IMAGE_TAG( 'o','u','t','l' ),
    'FT_GLYPH_FORMAT_PLOTTER'   : _FT_IMAGE_TAG( 'p','l','o','t' )}
globals().update(FT_GLYPH_FORMATS)
ft_glyph_format_none      = FT_GLYPH_FORMAT_NONE
ft_glyph_format_composite = FT_GLYPH_FORMAT_COMPOSITE
ft_glyph_format_bitmap    = FT_GLYPH_FORMAT_BITMAP
ft_glyph_format_outline   = FT_GLYPH_FORMAT_OUTLINE
ft_glyph_format_plotter   = FT_GLYPH_FORMAT_PLOTTER



# -----------------------------------------------------------------------------
# An enumeration used to specify character sets supported by charmaps. Used in
# the FT_Select_Charmap API function.
#
# FT_ENCODING_NONE
# The encoding value 0 is reserved.
#
# FT_ENCODING_UNICODE
# Corresponds to the Unicode character set. This value covers all versions of
# the Unicode repertoire, including ASCII and Latin-1. Most fonts include a
# Unicode charmap, but not all of them.
#
# For example, if you want to access Unicode value U+1F028 (and the font
# contains it), use value 0x1F028 as the input value for FT_Get_Char_Index.
#
# FT_ENCODING_MS_SYMBOL
# Corresponds to the Microsoft Symbol encoding, used to encode mathematical
# symbols in the 32..255 character code range. For more information, see
# 'http://www.ceviz.net/symbol.htm'.
#
# FT_ENCODING_SJIS
# Corresponds to Japanese SJIS encoding. More info at at
# 'http://langsupport.japanreference.com/encoding.shtml'. See note on
# multi-byte encodings below.
#
# FT_ENCODING_GB2312
# Corresponds to an encoding system for Simplified Chinese as used used in
# mainland China.
#
# FT_ENCODING_BIG5
# Corresponds to an encoding system for Traditional Chinese as used in Taiwan
# and Hong Kong.
#
# FT_ENCODING_WANSUNG
# Corresponds to the Korean encoding system known as Wansung. For more
# information see 'http://www.microsoft.com/typography/unicode/949.txt'.
#
# FT_ENCODING_JOHAB
# The Korean standard character set (KS C 5601-1992), which corresponds to MS
# Windows code page 1361. This character set includes all possible Hangeul
# character combinations.
#
# FT_ENCODING_ADOBE_LATIN_1
# Corresponds to a Latin-1 encoding as defined in a Type 1 PostScript font. It
# is limited to 256 character codes.
#
# FT_ENCODING_ADOBE_STANDARD
# Corresponds to the Adobe Standard encoding, as found in Type 1, CFF, and
# OpenType/CFF fonts. It is limited to 256 character codes.
#
# FT_ENCODING_ADOBE_EXPERT
# Corresponds to the Adobe Expert encoding, as found in Type 1, CFF, and
# OpenType/CFF fonts. It is limited to 256 character codes.
#
# FT_ENCODING_ADOBE_CUSTOM
# Corresponds to a custom encoding, as found in Type 1, CFF, and OpenType/CFF
# fonts. It is limited to 256 character codes.

# FT_ENCODING_APPLE_ROMAN
# Corresponds to the 8-bit Apple roman encoding. Many TrueType and OpenType
# fonts contain a charmap for this encoding, since older versions of Mac OS are
# able to use it.
#
# FT_ENCODING_OLD_LATIN_2
# This value is deprecated and was never used nor reported by FreeType. Don't
# use or test for it.
#
def _FT_ENC_TAG(a,b,c,d):
    return ( ord(a) << 24 | ord(b) << 16 | ord(c) << 8 | ord(d) )
FT_ENCODINGS = {'FT_ENCODING_NONE'           : _FT_ENC_TAG('\0','\0','\0','\0'),
                'FT_ENCODING_MS_SYMBOL'      : _FT_ENC_TAG( 's','y','m','b' ),
                'FT_ENCODING_UNICODE'        : _FT_ENC_TAG( 'u','n','i','c' ),
                'FT_ENCODING_SJIS'           : _FT_ENC_TAG( 's','j','i','s' ),
                'FT_ENCODING_GB2312'         : _FT_ENC_TAG( 'g','b',' ',' ' ),
                'FT_ENCODING_BIG5'           : _FT_ENC_TAG( 'b','i','g','5' ),
                'FT_ENCODING_WANSUNG'        : _FT_ENC_TAG( 'w','a','n','s' ),
                'FT_ENCODING_JOHAB'          : _FT_ENC_TAG( 'j','o','h','a' ),
                'FT_ENCODING_ADOBE_STANDARD' : _FT_ENC_TAG( 'A','D','O','B' ),
                'FT_ENCODING_ADOBE_EXPERT'   : _FT_ENC_TAG( 'A','D','B','E' ),
                'FT_ENCODING_ADOBE_CUSTOM'   : _FT_ENC_TAG( 'A','D','B','C' ),
                'FT_ENCODING_ADOBE_LATIN1'   : _FT_ENC_TAG( 'l','a','t','1' ),
                'FT_ENCODING_OLD_LATIN2'     : _FT_ENC_TAG( 'l','a','t','2' ),
                'FT_ENCODING_APPLE_ROMAN'    : _FT_ENC_TAG( 'a','r','m','n' ) }
globals().update(FT_ENCODINGS)



# -----------------------------------------------------------------------------
# An enumeration type that lists the render modes supported by FreeType 2. Each
# mode corresponds to a specific type of scanline conversion performed on the
# outline.
#
# For bitmap fonts and embedded bitmaps the 'bitmap->pixel_mode' field in the
# FT_GlyphSlotRec structure gives the format of the returned bitmap.
#
# All modes except FT_RENDER_MODE_MONO use 256 levels of opacity.
#
# FT_RENDER_MODE_NORMAL
# This is the default render mode; it corresponds to 8-bit anti-aliased
# bitmaps.
#
# FT_RENDER_MODE_LIGHT
# This is equivalent to FT_RENDER_MODE_NORMAL. It is only defined as a separate
# value because render modes are also used indirectly to define hinting
# algorithm selectors. See FT_LOAD_TARGET_XXX for details.
#
# FT_RENDER_MODE_MONO
# This mode corresponds to 1-bit bitmaps (with 2 levels of opacity).
#
# FT_RENDER_MODE_LCD
# This mode corresponds to horizontal RGB and BGR sub-pixel displays like LCD
# screens. It produces 8-bit bitmaps that are 3 times the width of the original
# glyph outline in pixels, and which use the FT_PIXEL_MODE_LCD mode.
#
# FT_RENDER_MODE_LCD_V
# This mode corresponds to vertical RGB and BGR sub-pixel displays (like PDA
# screens, rotated LCD displays, etc.). It produces 8-bit bitmaps that are 3
# times the height of the original glyph outline in pixels and use the
# FT_PIXEL_MODE_LCD_V mode.
#
FT_RENDER_MODES = { 'FT_RENDER_MODE_NORMAL' : 0,
                    'FT_RENDER_MODE_LIGHT'  : 1,
                    'FT_RENDER_MODE_MONO'   : 2,
                    'FT_RENDER_MODE_LCD'    : 3,
                    'FT_RENDER_MODE_LCD_V'  : 4 }
globals().update(FT_RENDER_MODES)



# -----------------------------------------------------------------------------
# A list of values that are used to select a specific hinting algorithm to use
# by the hinter. You should OR one of these values to your 'load_flags' when
# calling FT_Load_Glyph.
#
# Note that font's native hinters may ignore the hinting algorithm you have
# specified (e.g., the TrueType bytecode interpreter). You can set
# FT_LOAD_FORCE_AUTOHINT to ensure that the auto-hinter is used.
#
# Also note that FT_LOAD_TARGET_LIGHT is an exception, in that it always
# implies FT_LOAD_FORCE_AUTOHINT.
#
# FT_LOAD_TARGET_NORMAL
# This corresponds to the default hinting algorithm, optimized for standard
# gray-level rendering. For monochrome output, use FT_LOAD_TARGET_MONO instead.
#
# FT_LOAD_TARGET_LIGHT
# A lighter hinting algorithm for non-monochrome modes. Many generated glyphs
# are more fuzzy but better resemble its original shape. A bit like rendering
# on Mac OS X.
#
# As a special exception, this target implies FT_LOAD_FORCE_AUTOHINT.
#
# FT_LOAD_TARGET_MONO
# Strong hinting algorithm that should only be used for monochrome output. The
# result is probably unpleasant if the glyph is rendered in non-monochrome
# modes.
#
# FT_LOAD_TARGET_LCD
# A variant of FT_LOAD_TARGET_NORMAL optimized for horizontally decimated LCD
# displays.
#
# FT_LOAD_TARGET_LCD_V
# A variant of FT_LOAD_TARGET_NORMAL optimized for vertically decimated LCD
# displays.

def _FT_LOAD_TARGET_(x):
    return (x & 15) << 16
FT_LOAD_TARGETS = {
    'FT_LOAD_TARGET_NORMAL' : _FT_LOAD_TARGET_(FT_RENDER_MODE_NORMAL),
    'FT_LOAD_TARGET_LIGHT'  : _FT_LOAD_TARGET_(FT_RENDER_MODE_LIGHT),
    'FT_LOAD_TARGET_MONO'   : _FT_LOAD_TARGET_(FT_RENDER_MODE_MONO),
    'FT_LOAD_TARGET_LCD'    : _FT_LOAD_TARGET_(FT_RENDER_MODE_LCD),
    'FT_LOAD_TARGET_LCD_V'  : _FT_LOAD_TARGET_(FT_RENDER_MODE_LCD_V) }
globals().update(FT_LOAD_TARGETS)
#def FT_LOAD_TARGET_MODE(x):
#    return (x >> 16) & 15


# -----------------------------------------------------------------------------
# A list of bit-field constants used with FT_Load_Glyph to indicate what kind
# of operations to perform during glyph loading.
#
# FT_LOAD_DEFAULT
# Corresponding to 0, this value is used as the default glyph load
# operation. In this case, the following happens:
#
# 1. FreeType looks for a bitmap for the glyph corresponding to the face's
# current size. If one is found, the function returns. The bitmap data can be
# accessed from the glyph slot (see note below).
#
# 2. If no embedded bitmap is searched or found, FreeType looks for a scalable
# outline. If one is found, it is loaded from the font file, scaled to device
# pixels, then 'hinted' to the pixel grid in order to optimize it. The outline
# data can be accessed from the glyph slot (see note below).
#
# Note that by default, the glyph loader doesn't render outlines into
# bitmaps. The following flags are used to modify this default behaviour to
# more specific and useful cases.
#
# FT_LOAD_NO_SCALE
# Don't scale the outline glyph loaded, but keep it in font units.
#
# This flag implies FT_LOAD_NO_HINTING and FT_LOAD_NO_BITMAP, and unsets
# FT_LOAD_RENDER.
#
# FT_LOAD_NO_HINTING
# Disable hinting. This generally generates 'blurrier' bitmap glyph when the
# glyph is rendered in any of the anti-aliased modes. See also the note below.
#
# This flag is implied by FT_LOAD_NO_SCALE.
#
# FT_LOAD_RENDER
# Call FT_Render_Glyph after the glyph is loaded. By default, the glyph is
# rendered in FT_RENDER_MODE_NORMAL mode. This can be overridden by
# FT_LOAD_TARGET_XXX or FT_LOAD_MONOCHROME.
#
# This flag is unset by FT_LOAD_NO_SCALE.
#
# FT_LOAD_NO_BITMAP
# Ignore bitmap strikes when loading. Bitmap-only fonts ignore this flag.
#
# FT_LOAD_NO_SCALE always sets this flag.
#
# FT_LOAD_VERTICAL_LAYOUT
# Load the glyph for vertical text layout. Don't use it as it is problematic
# currently.
#
# FT_LOAD_FORCE_AUTOHINT
# Indicates that the auto-hinter is preferred over the font's native
# hinter. See also the note below.
#
# FT_LOAD_CROP_BITMAP
# Indicates that the font driver should crop the loaded bitmap glyph (i.e.,
# remove all space around its black bits). Not all drivers implement this.
#
# FT_LOAD_PEDANTIC
# Indicates that the font driver should perform pedantic verifications during
# glyph loading. This is mostly used to detect broken glyphs in fonts. By
# default, FreeType tries to handle broken fonts also.
#
# FT_LOAD_IGNORE_GLOBAL_ADVANCE_WIDTH
# Indicates that the font driver should ignore the global advance width defined
# in the font. By default, that value is used as the advance width for all
# glyphs when the face has FT_FACE_FLAG_FIXED_WIDTH set.
#
# This flag exists for historical reasons (to support buggy CJK fonts).
#
# FT_LOAD_NO_RECURSE
# This flag is only used internally. It merely indicates that the font driver
# should not load composite glyphs recursively. Instead, it should set the
# 'num_subglyph' and 'subglyphs' values of the glyph slot accordingly, and set
# 'glyph->format' to FT_GLYPH_FORMAT_COMPOSITE.
#
# The description of sub-glyphs is not available to client applications for now.
#
# This flag implies FT_LOAD_NO_SCALE and FT_LOAD_IGNORE_TRANSFORM.
#
# FT_LOAD_IGNORE_TRANSFORM
# Indicates that the transform matrix set by FT_Set_Transform should be ignored.
#
# FT_LOAD_MONOCHROME
# This flag is used with FT_LOAD_RENDER to indicate that you want to render an
# outline glyph to a 1-bit monochrome bitmap glyph, with 8 pixels packed into
# each byte of the bitmap data.
#
# Note that this has no effect on the hinting algorithm used. You should rather
# use FT_LOAD_TARGET_MONO so that the monochrome-optimized hinting algorithm is
# used.
#
# FT_LOAD_LINEAR_DESIGN
# Indicates that the 'linearHoriAdvance' and 'linearVertAdvance' fields of
# FT_GlyphSlotRec should be kept in font units. See FT_GlyphSlotRec for
# details.
#
# FT_LOAD_NO_AUTOHINT
# Disable auto-hinter. See also the note below.
#
FT_LOAD_FLAGS = { 'FT_LOAD_DEFAULT'                      : 0x0,
                  'FT_LOAD_NO_SCALE'                     : 0x1,
                  'FT_LOAD_NO_HINTING'                   : 0x2,
                  'FT_LOAD_RENDER'                       : 0x4,
                  'FT_LOAD_NO_BITMAP'                    : 0x8,
                  'FT_LOAD_VERTICAL_LAYOUT'              : 0x10,
                  'FT_LOAD_FORCE_AUTOHINT'               : 0x20,
                  'FT_LOAD_CROP_BITMAP'                  : 0x40,
                  'FT_LOAD_PEDANTIC'                     : 0x80,
                  'FT_LOAD_IGNORE_GLOBAL_ADVANCE_WIDTH'  : 0x200,
                  'FT_LOAD_NO_RECURSE'                   : 0x400,
                  'FT_LOAD_IGNORE_TRANSFORM'             : 0x800,
                  'FT_LOAD_MONOCHROME'                   : 0x1000,
                  'FT_LOAD_LINEAR_DESIGN'                : 0x2000,
                  'FT_LOAD_NO_AUTOHINT'                  : 0x8000 }
globals().update(FT_LOAD_FLAGS)



# -----------------------------------------------------------------------------
# A list of bit-flags used to indicate the style of a given face. These are
# used in the 'style_flags' field of FT_FaceRec.
#
# FT_STYLE_FLAG_ITALIC
# Indicates that a given face style is italic or oblique.
#
# FT_STYLE_FLAG_BOLD
# Indicates that a given face is bold.
#
FT_STYLE_FLAGS = {'FT_STYLE_FLAG_ITALIC' : 1,
                  'FT_STYLE_FLAG_BOLD'   : 2 }
globals().update(FT_STYLE_FLAGS)



# -----------------------------------------------------------------------------
# A list of bit flags used in the 'face_flags' field of the FT_FaceRec
# structure. They inform client applications of properties of the corresponding
# face.
#
# FT_FACE_FLAG_SCALABLE
# Indicates that the face contains outline glyphs. This doesn't prevent bitmap
# strikes, i.e., a face can have both this and and FT_FACE_FLAG_FIXED_SIZES
# set.
#
# FT_FACE_FLAG_FIXED_SIZES
# Indicates that the face contains bitmap strikes. See also the
# 'num_fixed_sizes' and 'available_sizes' fields of FT_FaceRec.
#
# FT_FACE_FLAG_FIXED_WIDTH
# Indicates that the face contains fixed-width characters (like Courier,
# Lucido, MonoType, etc.).
#
# FT_FACE_FLAG_SFNT
# Indicates that the face uses the 'sfnt' storage scheme. For now, this means
# TrueType and OpenType.
#
# FT_FACE_FLAG_HORIZONTAL
# Indicates that the face contains horizontal glyph metrics. This should be set
# for all common formats.
#
# FT_FACE_FLAG_VERTICAL
# Indicates that the face contains vertical glyph metrics. This is only
# available in some formats, not all of them.
#
# FT_FACE_FLAG_KERNING
# Indicates that the face contains kerning information. If set, the kerning
# distance can be retrieved through the function FT_Get_Kerning. Otherwise the
# function always return the vector (0,0). Note that FreeType doesn't handle
# kerning data from the 'GPOS' table (as present in some OpenType fonts).
#
#
# FT_FACE_FLAG_MULTIPLE_MASTERS
# Indicates that the font contains multiple masters and is capable of
# interpolating between them. See the multiple-masters specific API for
# details.
#
# FT_FACE_FLAG_GLYPH_NAMES
# Indicates that the font contains glyph names that can be retrieved through
# FT_Get_Glyph_Name. Note that some TrueType fonts contain broken glyph name
# tables. Use the function FT_Has_PS_Glyph_Names when needed.
#
# FT_FACE_FLAG_EXTERNAL_STREAM
# Used internally by FreeType to indicate that a face's stream was provided by
# the client application and should not be destroyed when FT_Done_Face is
# called. Don't read or test this flag.
#
# FT_FACE_FLAG_HINTER
# Set if the font driver has a hinting machine of its own. For example, with
# TrueType fonts, it makes sense to use data from the SFNT 'gasp' table only if
# the native TrueType hinting engine (with the bytecode interpreter) is
# available and active.
#
# FT_FACE_FLAG_CID_KEYED
# Set if the font is CID-keyed. In that case, the font is not accessed by glyph
# indices but by CID values. For subsetted CID-keyed fonts this has the
# consequence that not all index values are a valid argument to
# FT_Load_Glyph. Only the CID values for which corresponding glyphs in the
# subsetted font exist make FT_Load_Glyph return successfully; in all other
# cases you get an 'FT_Err_Invalid_Argument' error.
#
# Note that CID-keyed fonts which are in an SFNT wrapper don't have this flag
# set since the glyphs are accessed in the normal way (using contiguous
# indices); the 'CID-ness' isn't visible to the application.
#
# FT_FACE_FLAG_TRICKY
# Set if the font is 'tricky', this is, it always needs the font format's
# native hinting engine to get a reasonable result. A typical example is the
# Chinese font 'mingli.ttf' which uses TrueType bytecode instructions to move
# and scale all of its subglyphs.
#
# It is not possible to autohint such fonts using FT_LOAD_FORCE_AUTOHINT; it
# will also ignore FT_LOAD_NO_HINTING. You have to set both FT_LOAD_NO_HINTING
# and FT_LOAD_NO_AUTOHINT to really disable hinting; however, you probably
# never want this except for demonstration purposes.
#
# Currently, there are six TrueType fonts in the list of tricky fonts; they are
# hard-coded in file 'ttobjs.c'.
#
FT_FACE_FLAGS = { 'FT_FACE_FLAG_SCALABLE'          : 1 <<  0,
                  'FT_FACE_FLAG_FIXED_SIZES'       : 1 <<  1,
                  'FT_FACE_FLAG_FIXED_WIDTH'       : 1 <<  2,
                  'FT_FACE_FLAG_SFNT'              : 1 <<  3,
                  'FT_FACE_FLAG_HORIZONTAL'        : 1 <<  4,
                  'FT_FACE_FLAG_VERTICAL'          : 1 <<  5,
                  'FT_FACE_FLAG_KERNING'           : 1 <<  6,
                  'FT_FACE_FLAG_FAST_GLYPHS'       : 1 <<  7,
                  'FT_FACE_FLAG_MULTIPLE_MASTERS'  : 1 <<  8,
                  'FT_FACE_FLAG_GLYPH_NAMES'       : 1 <<  9,
                  'FT_FACE_FLAG_EXTERNAL_STREAM'   : 1 << 10,
                  'FT_FACE_FLAG_HINTER'            : 1 << 11,
                  'FT_FACE_FLAGS_CID_KEYED'        : 1 << 12,
                  'FT_FACE_FLAGS_CID_TRICKY'       : 1 << 13
 }
globals().update(FT_FACE_FLAGS)



# -----------------------------------------------------------------------------
# A list of bit-field constants use for the flags in an outline's 'flags'
# field.
#
# FT_OUTLINE_NONE
# Value 0 is reserved.
#
# FT_OUTLINE_OWNER
# If set, this flag indicates that the outline's field arrays (i.e., 'points',
# 'flags', and 'contours') are 'owned' by the outline object, and should thus
# be freed when it is destroyed.
#
# FT_OUTLINE_EVEN_ODD_FILL
# By default, outlines are filled using the non-zero winding rule. If set to 1,
# the outline will be filled using the even-odd fill rule (only works with the
# smooth rasterizer).
#
# FT_OUTLINE_REVERSE_FILL
# By default, outside contours of an outline are oriented in clock-wise
# direction, as defined in the TrueType specification. This flag is set if the
# outline uses the opposite direction (typically for Type 1 fonts). This flag
# is ignored by the scan converter.
#
# FT_OUTLINE_IGNORE_DROPOUTS
# By default, the scan converter will try to detect drop-outs in an outline and
# correct the glyph bitmap to ensure consistent shape continuity. If set, this
# flag hints the scan-line converter to ignore such cases. See below for more
# information.
#
# FT_OUTLINE_SMART_DROPOUTS
# Select smart dropout control. If unset, use simple dropout control. Ignored
# if FT_OUTLINE_IGNORE_DROPOUTS is set. See below for more information.
#
# FT_OUTLINE_INCLUDE_STUBS
# If set, turn pixels on for 'stubs', otherwise exclude them. Ignored if
# FT_OUTLINE_IGNORE_DROPOUTS is set. See below for more information.
#
# FT_OUTLINE_HIGH_PRECISION
# This flag indicates that the scan-line converter should try to convert this
# outline to bitmaps with the highest possible quality. It is typically set for
# small character sizes. Note that this is only a hint that might be completely
# ignored by a given scan-converter.
#
# FT_OUTLINE_SINGLE_PASS
# This flag is set to force a given scan-converter to only use a single pass
# over the outline to render a bitmap glyph image. Normally, it is set for very
# large character sizes. It is only a hint that might be completely ignored by
# a given scan-converter.
#
FT_OUTLINE_FLAGS = { 'FT_OUTLINE_NONE'            : 0x0,
                     'FT_OUTLINE_OWNER'           : 0x1,
                     'FT_OUTLINE_EVEN_ODD_FILL'   : 0x2,
                     'FT_OUTLINE_REVERSE_FILL'    : 0x4,
                     'FT_OUTLINE_IGNORE_DROPOUTS' : 0x8,
                     'FT_OUTLINE_SMART_DROPOUTS'  : 0x10,
                     'FT_OUTLINE_INCLUDE_STUBS'   : 0x20,
                     'FT_OUTLINE_HIGH_PRECISION'  : 0x100,
                     'FT_OUTLINE_SINGLE_PASS'     : 0x200 }
globals().update(FT_OUTLINE_FLAGS)



# -----------------------------------------------------------------------------
# A list of bit-field constants used within the 'flags' field of the
# FT_Open_Args structure.
#
# FT_OPEN_MEMORY
# This is a memory-based stream.
#
# FT_OPEN_STREAM
# Copy the stream from the 'stream' field.
#
# FT_OPEN_PATHNAME
# Create a new input stream from a C path name.
#
# FT_OPEN_DRIVER
# Use the 'driver' field.
#
# FT_OPEN_PARAMS
# Use the 'num_params' and 'params' fields.
#
FT_OPEN_MODES = {'FT_OPEN_MEMORY':   0x1,
                 'FT_OPEN_STREAM':   0x2,
                 'FT_OPEN_PATHNAME': 0x4,
                 'FT_OPEN_DRIVER':   0x8,
                 'FT_OPEN_PARAMS':   0x10 }
globals().update(FT_OPEN_MODES)



# -----------------------------------------------------------------------------
# An enumeration used to specify which kerning values to return in
# FT_Get_Kerning.
#
# FT_KERNING_DEFAULT
# Return scaled and grid-fitted kerning distances (value is 0).
#
# FT_KERNING_UNFITTED
# Return scaled but un-grid-fitted kerning distances.
#
# FT_KERNING_UNSCALED
# Return the kerning vector in original font units.
#
FT_KERNING_MODES = { 'FT_KERNING_DEFAULT'  : 0,
                     'FT_KERNING_UNFITTED' : 1,
                     'FT_KERNING_UNSCALED' : 2 }
globals().update(FT_KERNING_MODES)



# -----------------------------------------------------------------------------
# These values determine how two joining lines are rendered in a stroker.
#
# FT_STROKER_LINEJOIN_ROUND
# Used to render rounded line joins. Circular arcs are used to join two lines
# smoothly.
#
# FT_STROKER_LINEJOIN_BEVEL
# Used to render beveled line joins; i.e., the two joining lines are extended
# until they intersect.
#
# FT_STROKER_LINEJOIN_MITER
# Same as beveled rendering, except that an additional line break is added if
# the angle between the two joining lines is too closed (this is useful to
# avoid unpleasant spikes in beveled rendering).
#
FT_STROKER_LINEJOINS = { 'FT_STROKER_LINEJOIN_ROUND' : 0,
                         'FT_STROKER_LINEJOIN_BEVEL' : 1,
                         'FT_STROKER_LINEJOIN_MITER' : 2}
globals().update(FT_STROKER_LINEJOINS)



# -----------------------------------------------------------------------------
# These values determine how the end of opened sub-paths are rendered in a
# stroke.
#
# FT_STROKER_LINECAP_BUTT
# The end of lines is rendered as a full stop on the last point itself.
#
# FT_STROKER_LINECAP_ROUND
# The end of lines is rendered as a half-circle around the last point.
#
# FT_STROKER_LINECAP_SQUARE
# The end of lines is rendered as a square around the last point.
#
FT_STROKER_LINECAPS = { 'FT_STROKER_LINECAP_BUTT'   : 0,
                        'FT_STROKER_LINECAP_ROUND'  : 1,
                        'FT_STROKER_LINECAP_SQUARE' : 2}
globals().update(FT_STROKER_LINECAPS)



# -----------------------------------------------------------------------------
# These values are used to select a given stroke border in
# FT_Stroker_GetBorderCounts and FT_Stroker_ExportBorder.
#
# FT_STROKER_BORDER_LEFT
# Select the left border, relative to the drawing direction.
#
# FT_STROKER_BORDER_RIGHT
# Select the right border, relative to the drawing direction.
#
# Note
# Applications are generally interested in the 'inside' and 'outside'
# borders. However, there is no direct mapping between these and the 'left' and
# 'right' ones, since this really depends on the glyph's drawing orientation,
# which varies between font formats.
#
# You can however use FT_Outline_GetInsideBorder and
# FT_Outline_GetOutsideBorder to get these.
#
FT_STROKER_BORDERS = { 'FT_STROKER_BORDER_LEFT'  : 0,
                       'FT_STROKER_BORDER_RIGHT' : 1}
globals().update(FT_STROKER_BORDERS)



# -----------------------------------------------------------------------------
# A list of values to identify various types of LCD filters.
#
# FT_LCD_FILTER_NONE
# Do not perform filtering. When used with subpixel rendering, this results in
# sometimes severe color fringes.
#
# FT_LCD_FILTER_DEFAULT
# The default filter reduces color fringes considerably, at the cost of a
# slight blurriness in the output.
#
# FT_LCD_FILTER_LIGHT
# The light filter is a variant that produces less blurriness at the cost of
# slightly more color fringes than the default one. It might be better,
# depending on taste, your monitor, or your personal vision.
#
# FT_LCD_FILTER_LEGACY
# This filter corresponds to the original libXft color filter. It provides high
# contrast output but can exhibit really bad color fringes if glyphs are not
# extremely well hinted to the pixel grid. In other words, it only works well
# if the TrueType bytecode interpreter is enabled and high-quality hinted fonts
# are used.
#
# This filter is only provided for comparison purposes, and might be disabled
# or stay unsupported in the future.

FT_LCD_FILTERS = {'FT_LCD_FILTER_NONE'    : 0,
                  'FT_LCD_FILTER_DEFAULT' : 1,
                  'FT_LCD_FILTER_LIGHT'   : 2,
                  'FT_LCD_FILTER_LEGACY'  : 16}
globals().update(FT_LCD_FILTERS)



# -----------------------------------------------------------------------------
# A list of valid values for the 'platform_id' identifier code in FT_CharMapRec
# and FT_SfntName structures.
#
# TT_PLATFORM_APPLE_UNICODE
# Used by Apple to indicate a Unicode character map and/or name entry. See
# TT_APPLE_ID_XXX for corresponding 'encoding_id' values. Note that name
# entries in this format are coded as big-endian UCS-2 character codes only.
#
# TT_PLATFORM_MACINTOSH
# Used by Apple to indicate a MacOS-specific charmap and/or name entry. See
# TT_MAC_ID_XXX for corresponding 'encoding_id' values. Note that most TrueType
# fonts contain an Apple roman charmap to be usable on MacOS systems (even if
# they contain a Microsoft charmap as well).
#
# TT_PLATFORM_ISO
# This value was used to specify ISO/IEC 10646 charmaps. It is however now
# deprecated. See TT_ISO_ID_XXX for a list of corresponding 'encoding_id'
# values.
#
# TT_PLATFORM_MICROSOFT
# Used by Microsoft to indicate Windows-specific charmaps. See TT_MS_ID_XXX for
# a list of corresponding 'encoding_id' values. Note that most fonts contain a
# Unicode charmap using (TT_PLATFORM_MICROSOFT, TT_MS_ID_UNICODE_CS).
#
# TT_PLATFORM_CUSTOM
# Used to indicate application-specific charmaps.
#
# TT_PLATFORM_ADOBE
# This value isn't part of any font format specification, but is used by
# FreeType to report Adobe-specific charmaps in an FT_CharMapRec structure. See
# TT_ADOBE_ID_XXX.

TT_PLATFORMS = {
    'TT_PLATFORM_APPLE_UNICODE' : 0,
    'TT_PLATFORM_MACINTOSH'     : 1,
    'TT_PLATFORM_ISO'           : 2, # deprecated
    'TT_PLATFORM_MICROSOFT'     : 3,
    'TT_PLATFORM_CUSTOM'        : 4,
    'TT_PLATFORM_ADOBE'         : 7} # artificial
globals().update(TT_PLATFORMS)


# -----------------------------------------------------------------------------
# A list of valid values for the 'encoding_id' for TT_PLATFORM_APPLE_UNICODE
# charmaps and name entries.
#
# TT_APPLE_ID_DEFAULT
# Unicode version 1.0.
#
# TT_APPLE_ID_UNICODE_1_1
# Unicode 1.1; specifies Hangul characters starting at U+34xx.
#
# TT_APPLE_ID_ISO_10646
# Deprecated (identical to preceding).
#
# TT_APPLE_ID_UNICODE_2_0
# Unicode 2.0 and beyond (UTF-16 BMP only).
#
# TT_APPLE_ID_UNICODE_32
# Unicode 3.1 and beyond, using UTF-32.
#
# TT_APPLE_ID_VARIANT_SELECTOR
# From Adobe, not Apple. Not a normal cmap. Specifies variations on a real
# cmap.
TT_APPLE_IDS = {
    'TT_APPLE_ID_DEFAULT'          : 0,
    'TT_APPLE_ID_UNICODE_1_1'      : 1,
    'TT_APPLE_ID_ISO_10646'        : 2,
    'TT_APPLE_ID_UNICODE_2_0'      : 3,
    'TT_APPLE_ID_UNICODE_32'       : 4,
    'TT_APPLE_ID_VARIANT_SELECTOR' : 5 }
globals().update(TT_APPLE_IDS)



# -----------------------------------------------------------------------------
# A list of valid values for the 'encoding_id' for TT_PLATFORM_MACINTOSH
# charmaps and name entries.
TT_MAC_IDS = {
    'TT_MAC_ID_ROMAN'               :  0,
    'TT_MAC_ID_JAPANESE'            :  1,
    'TT_MAC_ID_TRADITIONAL_CHINESE' :  2,
    'TT_MAC_ID_KOREAN'              :  3,
    'TT_MAC_ID_ARABIC'              :  4,
    'TT_MAC_ID_HEBREW'              :  5,
    'TT_MAC_ID_GREEK'               :  6,
    'TT_MAC_ID_RUSSIAN'             :  7,
    'TT_MAC_ID_RSYMBOL'             :  8,
    'TT_MAC_ID_DEVANAGARI'          :  9,
    'TT_MAC_ID_GURMUKHI'            : 10,
    'TT_MAC_ID_GUJARATI'            : 11,
    'TT_MAC_ID_ORIYA'               : 12,
    'TT_MAC_ID_BENGALI'             : 13,
    'TT_MAC_ID_TAMIL'               : 14,
    'TT_MAC_ID_TELUGU'              : 15,
    'TT_MAC_ID_KANNADA'             : 16,
    'TT_MAC_ID_MALAYALAM'           : 17,
    'TT_MAC_ID_SINHALESE'           : 18,
    'TT_MAC_ID_BURMESE'             : 19,
    'TT_MAC_ID_KHMER'               : 20,
    'TT_MAC_ID_THAI'                : 21,
    'TT_MAC_ID_LAOTIAN'             : 22,
    'TT_MAC_ID_GEORGIAN'            : 23,
    'TT_MAC_ID_ARMENIAN'            : 24,
    'TT_MAC_ID_MALDIVIAN'           : 25,
    'TT_MAC_ID_SIMPLIFIED_CHINESE'  : 25,
    'TT_MAC_ID_TIBETAN'             : 26,
    'TT_MAC_ID_MONGOLIAN'           : 27,
    'TT_MAC_ID_GEEZ'                : 28,
    'TT_MAC_ID_SLAVIC'              : 29,
    'TT_MAC_ID_VIETNAMESE'          : 30,
    'TT_MAC_ID_SINDHI'              : 31,
    'TT_MAC_ID_UNINTERP'            : 32}
globals().update(TT_MAC_IDS)


# -----------------------------------------------------------------------------
# A list of valid values for the 'encoding_id' for TT_PLATFORM_MICROSOFT
# charmaps and name entries.
#
# TT_MS_ID_SYMBOL_CS
# Corresponds to Microsoft symbol encoding. See FT_ENCODING_MS_SYMBOL.
#
# TT_MS_ID_UNICODE_CS
# Corresponds to a Microsoft WGL4 charmap, matching Unicode. See
# FT_ENCODING_UNICODE.
#
# TT_MS_ID_SJIS
# Corresponds to SJIS Japanese encoding. See FT_ENCODING_SJIS.
#
# TT_MS_ID_GB2312
# Corresponds to Simplified Chinese as used in Mainland China. See
# FT_ENCODING_GB2312.
#
# TT_MS_ID_BIG_5
# Corresponds to Traditional Chinese as used in Taiwan and Hong Kong. See
# FT_ENCODING_BIG5.
#
# TT_MS_ID_WANSUNG
# Corresponds to Korean Wansung encoding. See FT_ENCODING_WANSUNG.
#
# TT_MS_ID_JOHAB
# Corresponds to Johab encoding. See FT_ENCODING_JOHAB.
#
# TT_MS_ID_UCS_4
# Corresponds to UCS-4 or UTF-32 charmaps. This has been added to the OpenType
# specification version 1.4 (mid-2001.)

TT_MS_IDS = {
    'TT_MS_ID_SYMBOL_CS'  :  0,
    'TT_MS_ID_UNICODE_CS' :  1,
    'TT_MS_ID_SJIS'       :  2,
    'TT_MS_ID_GB2312'     :  3,
    'TT_MS_ID_BIG_5'      :  4,
    'TT_MS_ID_WANSUNG'    :  5,
    'TT_MS_ID_JOHAB'      :  6,
    'TT_MS_ID_UCS_4'      : 10 }
globals().update(TT_MS_IDS)


# -----------------------------------------------------------------------------
# A list of valid values for the 'encoding_id' for TT_PLATFORM_ADOBE
# charmaps. This is a FreeType-specific extension!
#
# TT_ADOBE_ID_STANDARD
# Adobe standard encoding.
#
# TT_ADOBE_ID_EXPERT
# Adobe expert encoding.
#
# TT_ADOBE_ID_CUSTOM
# Adobe custom encoding.
#
# TT_ADOBE_ID_LATIN_1
# Adobe Latin 1 encoding.

TT_ADOBE_IDS = {
    'TT_ADOBE_ID_STANDARD' : 0,
    'TT_ADOBE_ID_EXPERT'   : 1,
    'TT_ADOBE_ID_CUSTOM'   : 2,
    'TT_ADOBE_ID_LATIN_1'  : 3 }
globals().update(TT_ADOBE_IDS)


# -----------------------------------------------------------------------------
#  Possible values of the language identifier field in the name records of the
#  TTF `name' table if the `platform' identifier code is TT_PLATFORM_MACINTOSH.
TT_MAC_LANGIDS = {
    'TT_MAC_LANGID_ENGLISH'                    :   0,
    'TT_MAC_LANGID_FRENCH'                     :   1,
    'TT_MAC_LANGID_GERMAN'                     :   2,
    'TT_MAC_LANGID_ITALIAN'                    :   3,
    'TT_MAC_LANGID_DUTCH'                      :   4,
    'TT_MAC_LANGID_SWEDISH'                    :   5,
    'TT_MAC_LANGID_SPANISH'                    :   6,
    'TT_MAC_LANGID_DANISH'                     :   7,
    'TT_MAC_LANGID_PORTUGUESE'                 :   8,
    'TT_MAC_LANGID_NORWEGIAN'                  :   9,
    'TT_MAC_LANGID_HEBREW'                     :  10,
    'TT_MAC_LANGID_JAPANESE'                   :  11,
    'TT_MAC_LANGID_ARABIC'                     :  12,
    'TT_MAC_LANGID_FINNISH'                    :  13,
    'TT_MAC_LANGID_GREEK'                      :  14,
    'TT_MAC_LANGID_ICELANDIC'                  :  15,
    'TT_MAC_LANGID_MALTESE'                    :  16,
    'TT_MAC_LANGID_TURKISH'                    :  17,
    'TT_MAC_LANGID_CROATIAN'                   :  18,
    'TT_MAC_LANGID_CHINESE_TRADITIONAL'        :  19,
    'TT_MAC_LANGID_URDU'                       :  20,
    'TT_MAC_LANGID_HINDI'                      :  21,
    'TT_MAC_LANGID_THAI'                       :  22,
    'TT_MAC_LANGID_KOREAN'                     :  23,
    'TT_MAC_LANGID_LITHUANIAN'                 :  24,
    'TT_MAC_LANGID_POLISH'                     :  25,
    'TT_MAC_LANGID_HUNGARIAN'                  :  26,
    'TT_MAC_LANGID_ESTONIAN'                   :  27,
    'TT_MAC_LANGID_LETTISH'                    :  28,
    'TT_MAC_LANGID_SAAMISK'                    :  29,
    'TT_MAC_LANGID_FAEROESE'                   :  30,
    'TT_MAC_LANGID_FARSI'                      :  31,
    'TT_MAC_LANGID_RUSSIAN'                    :  32,
    'TT_MAC_LANGID_CHINESE_SIMPLIFIED'         :  33,
    'TT_MAC_LANGID_FLEMISH'                    :  34,
    'TT_MAC_LANGID_IRISH'                      :  35,
    'TT_MAC_LANGID_ALBANIAN'                   :  36,
    'TT_MAC_LANGID_ROMANIAN'                   :  37,
    'TT_MAC_LANGID_CZECH'                      :  38,
    'TT_MAC_LANGID_SLOVAK'                     :  39,
    'TT_MAC_LANGID_SLOVENIAN'                  :  40,
    'TT_MAC_LANGID_YIDDISH'                    :  41,
    'TT_MAC_LANGID_SERBIAN'                    :  42,
    'TT_MAC_LANGID_MACEDONIAN'                 :  43,
    'TT_MAC_LANGID_BULGARIAN'                  :  44,
    'TT_MAC_LANGID_UKRAINIAN'                  :  45,
    'TT_MAC_LANGID_BYELORUSSIAN'               :  46,
    'TT_MAC_LANGID_UZBEK'                      :  47,
    'TT_MAC_LANGID_KAZAKH'                     :  48,
    'TT_MAC_LANGID_AZERBAIJANI'                :  49,
    'TT_MAC_LANGID_AZERBAIJANI_CYRILLIC_SCRIPT':  49,
    'TT_MAC_LANGID_AZERBAIJANI_ARABIC_SCRIPT'  :  50,
    'TT_MAC_LANGID_ARMENIAN'                   :  51,
    'TT_MAC_LANGID_GEORGIAN'                   :  52,
    'TT_MAC_LANGID_MOLDAVIAN'                  :  53,
    'TT_MAC_LANGID_KIRGHIZ'                    :  54,
    'TT_MAC_LANGID_TAJIKI'                     :  55,
    'TT_MAC_LANGID_TURKMEN'                    :  56,
    'TT_MAC_LANGID_MONGOLIAN'                  :  57,
    'TT_MAC_LANGID_MONGOLIAN_MONGOLIAN_SCRIPT' :  57,
    'TT_MAC_LANGID_MONGOLIAN_CYRILLIC_SCRIPT'  :  58,
    'TT_MAC_LANGID_PASHTO'                     :  59,
    'TT_MAC_LANGID_KURDISH'                    :  60,
    'TT_MAC_LANGID_KASHMIRI'                   :  61,
    'TT_MAC_LANGID_SINDHI'                     :  62,
    'TT_MAC_LANGID_TIBETAN'                    :  63,
    'TT_MAC_LANGID_NEPALI'                     :  64,
    'TT_MAC_LANGID_SANSKRIT'                   :  65,
    'TT_MAC_LANGID_MARATHI'                    :  66,
    'TT_MAC_LANGID_BENGALI'                    :  67,
    'TT_MAC_LANGID_ASSAMESE'                   :  68,
    'TT_MAC_LANGID_GUJARATI'                   :  69,
    'TT_MAC_LANGID_PUNJABI'                    :  70,
    'TT_MAC_LANGID_ORIYA'                      :  71,
    'TT_MAC_LANGID_MALAYALAM'                  :  72,
    'TT_MAC_LANGID_KANNADA'                    :  73,
    'TT_MAC_LANGID_TAMIL'                      :  74,
    'TT_MAC_LANGID_TELUGU'                     :  75,
    'TT_MAC_LANGID_SINHALESE'                  :  76,
    'TT_MAC_LANGID_BURMESE'                    :  77,
    'TT_MAC_LANGID_KHMER'                      :  78,
    'TT_MAC_LANGID_LAO'                        :  79,
    'TT_MAC_LANGID_VIETNAMESE'                 :  80,
    'TT_MAC_LANGID_INDONESIAN'                 :  81,
    'TT_MAC_LANGID_TAGALOG'                    :  82,
    'TT_MAC_LANGID_MALAY_ROMAN_SCRIPT'         :  83,
    'TT_MAC_LANGID_MALAY_ARABIC_SCRIPT'        :  84,
    'TT_MAC_LANGID_AMHARIC'                    :  85,
    'TT_MAC_LANGID_TIGRINYA'                   :  86,
    'TT_MAC_LANGID_GALLA'                      :  87,
    'TT_MAC_LANGID_SOMALI'                     :  88,
    'TT_MAC_LANGID_SWAHILI'                    :  89,
    'TT_MAC_LANGID_RUANDA'                     :  90,
    'TT_MAC_LANGID_RUNDI'                      :  91,
    'TT_MAC_LANGID_CHEWA'                      :  92,
    'TT_MAC_LANGID_MALAGASY'                   :  93,
    'TT_MAC_LANGID_ESPERANTO'                  :  94,
    'TT_MAC_LANGID_WELSH'                      : 128,
    'TT_MAC_LANGID_BASQUE'                     : 129,
    'TT_MAC_LANGID_CATALAN'                    : 130,
    'TT_MAC_LANGID_LATIN'                      : 131,
    'TT_MAC_LANGID_QUECHUA'                    : 132,
    'TT_MAC_LANGID_GUARANI'                    : 133,
    'TT_MAC_LANGID_AYMARA'                     : 134,
    'TT_MAC_LANGID_TATAR'                      : 135,
    'TT_MAC_LANGID_UIGHUR'                     : 136,
    'TT_MAC_LANGID_DZONGKHA'                   : 137,
    'TT_MAC_LANGID_JAVANESE'                   : 138,
    'TT_MAC_LANGID_SUNDANESE'                  : 139,
    'TT_MAC_LANGID_GALICIAN'                   : 140,
    'TT_MAC_LANGID_AFRIKAANS'                  : 141,
    'TT_MAC_LANGID_BRETON'                     : 142,
    'TT_MAC_LANGID_INUKTITUT'                  : 143,
    'TT_MAC_LANGID_SCOTTISH_GAELIC'            : 144,
    'TT_MAC_LANGID_MANX_GAELIC'                : 145,
    'TT_MAC_LANGID_IRISH_GAELIC'               : 146,
    'TT_MAC_LANGID_TONGAN'                     : 147,
    'TT_MAC_LANGID_GREEK_POLYTONIC'            : 148,
    'TT_MAC_LANGID_GREELANDIC'                 : 149,
    'TT_MAC_LANGID_AZERBAIJANI_ROMAN_SCRIPT'   : 150 }
globals().update(TT_MAC_LANGIDS)



# -----------------------------------------------------------------------------
# Possible values of the language identifier field in the name records of the
# TTF `name' table if the `platform' identifier code is TT_PLATFORM_MICROSOFT.

TT_MS_LANGIDS = {
    'TT_MS_LANGID_ARABIC_GENERAL'                    : 0x0001,
    'TT_MS_LANGID_ARABIC_SAUDI_ARABIA'               : 0x0401,
    'TT_MS_LANGID_ARABIC_IRAQ'                       : 0x0801,
    'TT_MS_LANGID_ARABIC_EGYPT'                      : 0x0c01,
    'TT_MS_LANGID_ARABIC_LIBYA'                      : 0x1001,
    'TT_MS_LANGID_ARABIC_ALGERIA'                    : 0x1401,
    'TT_MS_LANGID_ARABIC_MOROCCO'                    : 0x1801,
    'TT_MS_LANGID_ARABIC_TUNISIA'                    : 0x1c01,
    'TT_MS_LANGID_ARABIC_OMAN'                       : 0x2001,
    'TT_MS_LANGID_ARABIC_YEMEN'                      : 0x2401,
    'TT_MS_LANGID_ARABIC_SYRIA'                      : 0x2801,
    'TT_MS_LANGID_ARABIC_JORDAN'                     : 0x2c01,
    'TT_MS_LANGID_ARABIC_LEBANON'                    : 0x3001,
    'TT_MS_LANGID_ARABIC_KUWAIT'                     : 0x3401,
    'TT_MS_LANGID_ARABIC_UAE'                        : 0x3801,
    'TT_MS_LANGID_ARABIC_BAHRAIN'                    : 0x3c01,
    'TT_MS_LANGID_ARABIC_QATAR'                      : 0x4001,
    'TT_MS_LANGID_BULGARIAN_BULGARIA'                : 0x0402,
    'TT_MS_LANGID_CATALAN_SPAIN'                     : 0x0403,
    'TT_MS_LANGID_CHINESE_GENERAL'                   : 0x0004,
    'TT_MS_LANGID_CHINESE_TAIWAN'                    : 0x0404,
    'TT_MS_LANGID_CHINESE_PRC'                       : 0x0804,
    'TT_MS_LANGID_CHINESE_HONG_KONG'                 : 0x0c04,
    'TT_MS_LANGID_CHINESE_SINGAPORE'                 : 0x1004,
    'TT_MS_LANGID_CHINESE_MACAU'                     : 0x1404,
    'TT_MS_LANGID_CZECH_CZECH_REPUBLIC'              : 0x0405,
    'TT_MS_LANGID_DANISH_DENMARK'                    : 0x0406,
    'TT_MS_LANGID_GERMAN_GERMANY'                    : 0x0407,
    'TT_MS_LANGID_GERMAN_SWITZERLAND'                : 0x0807,
    'TT_MS_LANGID_GERMAN_AUSTRIA'                    : 0x0c07,
    'TT_MS_LANGID_GERMAN_LUXEMBOURG'                 : 0x1007,
    'TT_MS_LANGID_GERMAN_LIECHTENSTEI'               : 0x1407,
    'TT_MS_LANGID_GREEK_GREECE'                      : 0x0408,
    'TT_MS_LANGID_ENGLISH_GENERAL'                   : 0x0009,
    'TT_MS_LANGID_ENGLISH_UNITED_STATES'             : 0x0409,
    'TT_MS_LANGID_ENGLISH_UNITED_KINGDOM'            : 0x0809,
    'TT_MS_LANGID_ENGLISH_AUSTRALIA'                 : 0x0c09,
    'TT_MS_LANGID_ENGLISH_CANADA'                    : 0x1009,
    'TT_MS_LANGID_ENGLISH_NEW_ZEALAND'               : 0x1409,
    'TT_MS_LANGID_ENGLISH_IRELAND'                   : 0x1809,
    'TT_MS_LANGID_ENGLISH_SOUTH_AFRICA'              : 0x1c09,
    'TT_MS_LANGID_ENGLISH_JAMAICA'                   : 0x2009,
    'TT_MS_LANGID_ENGLISH_CARIBBEAN'                 : 0x2409,
    'TT_MS_LANGID_ENGLISH_BELIZE'                    : 0x2809,
    'TT_MS_LANGID_ENGLISH_TRINIDAD'                  : 0x2c09,
    'TT_MS_LANGID_ENGLISH_ZIMBABWE'                  : 0x3009,
    'TT_MS_LANGID_ENGLISH_PHILIPPINES'               : 0x3409,
    'TT_MS_LANGID_ENGLISH_INDONESIA'                 : 0x3809,
    'TT_MS_LANGID_ENGLISH_HONG_KONG'                 : 0x3c09,
    'TT_MS_LANGID_ENGLISH_INDIA'                     : 0x4009,
    'TT_MS_LANGID_ENGLISH_MALAYSIA'                  : 0x4409,
    'TT_MS_LANGID_ENGLISH_SINGAPORE'                 : 0x4809,
    'TT_MS_LANGID_SPANISH_SPAIN_TRADITIONAL_SORT'    : 0x040a,
    'TT_MS_LANGID_SPANISH_MEXICO'                    : 0x080a,
    'TT_MS_LANGID_SPANISH_SPAIN_INTERNATIONAL_SORT'  : 0x0c0a,
    'TT_MS_LANGID_SPANISH_GUATEMALA'                 : 0x100a,
    'TT_MS_LANGID_SPANISH_COSTA_RICA'                : 0x140a,
    'TT_MS_LANGID_SPANISH_PANAMA'                    : 0x180a,
    'TT_MS_LANGID_SPANISH_DOMINICAN_REPUBLIC'        : 0x1c0a,
    'TT_MS_LANGID_SPANISH_VENEZUELA'                 : 0x200a,
    'TT_MS_LANGID_SPANISH_COLOMBIA'                  : 0x240a,
    'TT_MS_LANGID_SPANISH_PERU'                      : 0x280a,
    'TT_MS_LANGID_SPANISH_ARGENTINA'                 : 0x2c0a,
    'TT_MS_LANGID_SPANISH_ECUADOR'                   : 0x300a,
    'TT_MS_LANGID_SPANISH_CHILE'                     : 0x340a,
    'TT_MS_LANGID_SPANISH_URUGUAY'                   : 0x380a,
    'TT_MS_LANGID_SPANISH_PARAGUAY'                  : 0x3c0a,
    'TT_MS_LANGID_SPANISH_BOLIVIA'                   : 0x400a,
    'TT_MS_LANGID_SPANISH_EL_SALVADOR'               : 0x440a,
    'TT_MS_LANGID_SPANISH_HONDURAS'                  : 0x480a,
    'TT_MS_LANGID_SPANISH_NICARAGUA'                 : 0x4c0a,
    'TT_MS_LANGID_SPANISH_PUERTO_RICO'               : 0x500a,
    'TT_MS_LANGID_SPANISH_UNITED_STATES'             : 0x540a,
    'TT_MS_LANGID_SPANISH_LATIN_AMERICA'             : 0xE40a,
    'TT_MS_LANGID_FINNISH_FINLAND'                   : 0x040b,
    'TT_MS_LANGID_FRENCH_FRANCE'                     : 0x040c,
    'TT_MS_LANGID_FRENCH_BELGIUM'                    : 0x080c,
    'TT_MS_LANGID_FRENCH_CANADA'                     : 0x0c0c,
    'TT_MS_LANGID_FRENCH_SWITZERLAND'                : 0x100c,
    'TT_MS_LANGID_FRENCH_LUXEMBOURG'                 : 0x140c,
    'TT_MS_LANGID_FRENCH_MONACO'                     : 0x180c,
    'TT_MS_LANGID_FRENCH_WEST_INDIES'                : 0x1c0c,
    'TT_MS_LANGID_FRENCH_REUNION'                    : 0x200c,
    'TT_MS_LANGID_FRENCH_CONGO'                      : 0x240c,
    'TT_MS_LANGID_FRENCH_SENEGAL'                    : 0x280c,
    'TT_MS_LANGID_FRENCH_CAMEROON'                   : 0x2c0c,
    'TT_MS_LANGID_FRENCH_COTE_D_IVOIRE'              : 0x300c,
    'TT_MS_LANGID_FRENCH_MALI'                       : 0x340c,
    'TT_MS_LANGID_FRENCH_MOROCCO'                    : 0x380c,
    'TT_MS_LANGID_FRENCH_HAITI'                      : 0x3c0c,
    'TT_MS_LANGID_FRENCH_NORTH_AFRICA'               : 0xE40c,
    'TT_MS_LANGID_HEBREW_ISRAEL'                     : 0x040d,
    'TT_MS_LANGID_HUNGARIAN_HUNGARY'                 : 0x040e,
    'TT_MS_LANGID_ICELANDIC_ICELAND'                 : 0x040f,
    'TT_MS_LANGID_ITALIAN_ITALY'                     : 0x0410,
    'TT_MS_LANGID_ITALIAN_SWITZERLAND'               : 0x0810,
    'TT_MS_LANGID_JAPANESE_JAPAN'                    : 0x0411,
    'TT_MS_LANGID_KOREAN_EXTENDED_WANSUNG_KOREA'     : 0x0412,
    'TT_MS_LANGID_KOREAN_JOHAB_KOREA'                : 0x0812,
    'TT_MS_LANGID_DUTCH_NETHERLANDS'                 : 0x0413,
    'TT_MS_LANGID_DUTCH_BELGIUM'                     : 0x0813,
    'TT_MS_LANGID_NORWEGIAN_NORWAY_BOKMAL'           : 0x0414,
    'TT_MS_LANGID_NORWEGIAN_NORWAY_NYNORSK'          : 0x0814,
    'TT_MS_LANGID_POLISH_POLAND'                     : 0x0415,
    'TT_MS_LANGID_PORTUGUESE_BRAZIL'                 : 0x0416,
    'TT_MS_LANGID_PORTUGUESE_PORTUGAL'               : 0x0816,
    'TT_MS_LANGID_RHAETO_ROMANIC_SWITZERLAND'        : 0x0417,
    'TT_MS_LANGID_ROMANIAN_ROMANIA'                  : 0x0418,
    'TT_MS_LANGID_MOLDAVIAN_MOLDAVIA'                : 0x0818,
    'TT_MS_LANGID_RUSSIAN_RUSSIA'                    : 0x0419,
    'TT_MS_LANGID_RUSSIAN_MOLDAVIA'                  : 0x0819,
    'TT_MS_LANGID_CROATIAN_CROATIA'                  : 0x041a,
    'TT_MS_LANGID_SERBIAN_SERBIA_LATIN'              : 0x081a,
    'TT_MS_LANGID_SERBIAN_SERBIA_CYRILLIC'           : 0x0c1a,
    'TT_MS_LANGID_CROATIAN_BOSNIA_HERZEGOVINA'       : 0x101a,
    'TT_MS_LANGID_BOSNIAN_BOSNIA_HERZEGOVINA'        : 0x141a,
    'TT_MS_LANGID_SERBIAN_BOSNIA_HERZ_LATIN'         : 0x181a,
    'TT_MS_LANGID_SERBIAN_BOSNIA_HERZ_CYRILLIC'      : 0x181a,
    'TT_MS_LANGID_SLOVAK_SLOVAKIA'                   : 0x041b,
    'TT_MS_LANGID_ALBANIAN_ALBANIA'                  : 0x041c,
    'TT_MS_LANGID_SWEDISH_SWEDEN'                    : 0x041d,
    'TT_MS_LANGID_SWEDISH_FINLAND'                   : 0x081d,
    'TT_MS_LANGID_THAI_THAILAND'                     : 0x041e,
    'TT_MS_LANGID_TURKISH_TURKEY'                    : 0x041f,
    'TT_MS_LANGID_URDU_PAKISTAN'                     : 0x0420,
    'TT_MS_LANGID_URDU_INDIA'                        : 0x0820,
    'TT_MS_LANGID_INDONESIAN_INDONESIA'              : 0x0421,
    'TT_MS_LANGID_UKRAINIAN_UKRAINE'                 : 0x0422,
    'TT_MS_LANGID_BELARUSIAN_BELARUS'                : 0x0423,
    'TT_MS_LANGID_SLOVENE_SLOVENIA'                  : 0x0424,
    'TT_MS_LANGID_ESTONIAN_ESTONIA'                  : 0x0425,
    'TT_MS_LANGID_LATVIAN_LATVIA'                    : 0x0426,
    'TT_MS_LANGID_LITHUANIAN_LITHUANIA'              : 0x0427,
    'TT_MS_LANGID_CLASSIC_LITHUANIAN_LITHUANIA'      : 0x0827,
    'TT_MS_LANGID_TAJIK_TAJIKISTAN'                  : 0x0428,
    'TT_MS_LANGID_FARSI_IRAN'                        : 0x0429,
    'TT_MS_LANGID_VIETNAMESE_VIET_NAM'               : 0x042a,
    'TT_MS_LANGID_ARMENIAN_ARMENIA'                  : 0x042b,
    'TT_MS_LANGID_AZERI_AZERBAIJAN_LATIN'            : 0x042c,
    'TT_MS_LANGID_AZERI_AZERBAIJAN_CYRILLIC'         : 0x082c,
    'TT_MS_LANGID_BASQUE_SPAIN'                      : 0x042d,
    'TT_MS_LANGID_SORBIAN_GERMANY'                   : 0x042e,
    'TT_MS_LANGID_MACEDONIAN_MACEDONIA'              : 0x042f,
    'TT_MS_LANGID_SUTU_SOUTH_AFRICA'                 : 0x0430,
    'TT_MS_LANGID_TSONGA_SOUTH_AFRICA'               : 0x0431,
    'TT_MS_LANGID_TSWANA_SOUTH_AFRICA'               : 0x0432,
    'TT_MS_LANGID_VENDA_SOUTH_AFRICA'                : 0x0433,
    'TT_MS_LANGID_XHOSA_SOUTH_AFRICA'                : 0x0434,
    'TT_MS_LANGID_ZULU_SOUTH_AFRICA'                 : 0x0435,
    'TT_MS_LANGID_AFRIKAANS_SOUTH_AFRICA'            : 0x0436,
    'TT_MS_LANGID_GEORGIAN_GEORGIA'                  : 0x0437,
    'TT_MS_LANGID_FAEROESE_FAEROE_ISLANDS'           : 0x0438,
    'TT_MS_LANGID_HINDI_INDIA'                       : 0x0439,
    'TT_MS_LANGID_MALTESE_MALTA'                     : 0x043a,
    'TT_MS_LANGID_SAMI_NORTHERN_NORWAY'              : 0x043b,
    'TT_MS_LANGID_SAMI_NORTHERN_SWEDEN'              : 0x083b,
    'TT_MS_LANGID_SAMI_NORTHERN_FINLAND'             : 0x0C3b,
    'TT_MS_LANGID_SAMI_LULE_NORWAY'                  : 0x103b,
    'TT_MS_LANGID_SAMI_LULE_SWEDEN'                  : 0x143b,
    'TT_MS_LANGID_SAMI_SOUTHERN_NORWAY'              : 0x183b,
    'TT_MS_LANGID_SAMI_SOUTHERN_SWEDEN'              : 0x1C3b,
    'TT_MS_LANGID_SAMI_SKOLT_FINLAND'                : 0x203b,
    'TT_MS_LANGID_SAMI_INARI_FINLAND'                : 0x243b,
    'TT_MS_LANGID_SAAMI_LAPONIA'                     : 0x043b,
    'TT_MS_LANGID_SCOTTISH_GAELIC_UNITED_KINGDOM'    : 0x083c,
    'TT_MS_LANGID_IRISH_GAELIC_IRELAND'              : 0x043c,
    'TT_MS_LANGID_YIDDISH_GERMANY'                   : 0x043d,
    'TT_MS_LANGID_MALAY_MALAYSIA'                    : 0x043e,
    'TT_MS_LANGID_MALAY_BRUNEI_DARUSSALAM'           : 0x083e,
    'TT_MS_LANGID_KAZAK_KAZAKSTAN'                   : 0x043f,
    'TT_MS_LANGID_KIRGHIZ_KIRGHIZSTAN'               : 0x0440,
    'TT_MS_LANGID_KIRGHIZ_KIRGHIZ_REPUBLIC'          : 0x0440,
    'TT_MS_LANGID_SWAHILI_KENYA'                     : 0x0441,
    'TT_MS_LANGID_TURKMEN_TURKMENISTAN'              : 0x0442,
    'TT_MS_LANGID_UZBEK_UZBEKISTAN_LATIN'            : 0x0443,
    'TT_MS_LANGID_UZBEK_UZBEKISTAN_CYRILLIC'         : 0x0843,
    'TT_MS_LANGID_TATAR_TATARSTAN'                   : 0x0444,
    'TT_MS_LANGID_BENGALI_INDIA'                     : 0x0445,
    'TT_MS_LANGID_BENGALI_BANGLADESH'                : 0x0845,
    'TT_MS_LANGID_PUNJABI_INDIA'                     : 0x0446,
    'TT_MS_LANGID_PUNJABI_ARABIC_PAKISTAN'           : 0x0846,
    'TT_MS_LANGID_GUJARATI_INDIA'                    : 0x0447,
    'TT_MS_LANGID_ORIYA_INDIA'                       : 0x0448,
    'TT_MS_LANGID_TAMIL_INDIA'                       : 0x0449,
    'TT_MS_LANGID_TELUGU_INDIA'                      : 0x044a,
    'TT_MS_LANGID_KANNADA_INDIA'                     : 0x044b,
    'TT_MS_LANGID_MALAYALAM_INDIA'                   : 0x044c,
    'TT_MS_LANGID_ASSAMESE_INDIA'                    : 0x044d,
    'TT_MS_LANGID_MARATHI_INDIA'                     : 0x044e,
    'TT_MS_LANGID_SANSKRIT_INDIA'                    : 0x044f,
    'TT_MS_LANGID_MONGOLIAN_MONGOLIA'                : 0x0450,
    'TT_MS_LANGID_MONGOLIAN_MONGOLIA_MONGOLIAN'      : 0x0850,
    'TT_MS_LANGID_TIBETAN_CHINA'                     : 0x0451,
    'TT_MS_LANGID_DZONGHKA_BHUTAN'                   : 0x0851,
    'TT_MS_LANGID_TIBETAN_BHUTAN'                    : 0x0851,
    'TT_MS_LANGID_WELSH_WALES'                       : 0x0452,
    'TT_MS_LANGID_KHMER_CAMBODIA'                    : 0x0453,
    'TT_MS_LANGID_LAO_LAOS'                          : 0x0454,
    'TT_MS_LANGID_BURMESE_MYANMAR'                   : 0x0455,
    'TT_MS_LANGID_GALICIAN_SPAIN'                    : 0x0456,
    'TT_MS_LANGID_KONKANI_INDIA'                     : 0x0457,
    'TT_MS_LANGID_MANIPURI_INDIA'                    : 0x0458,
    'TT_MS_LANGID_SINDHI_INDIA'                      : 0x0459,
    'TT_MS_LANGID_SINDHI_PAKISTAN'                   : 0x0859,
    'TT_MS_LANGID_SYRIAC_SYRIA'                      : 0x045a,
    'TT_MS_LANGID_SINHALESE_SRI_LANKA'               : 0x045b,
    'TT_MS_LANGID_CHEROKEE_UNITED_STATES'            : 0x045c,
    'TT_MS_LANGID_INUKTITUT_CANADA'                  : 0x045d,
    'TT_MS_LANGID_AMHARIC_ETHIOPIA'                  : 0x045e,
    'TT_MS_LANGID_TAMAZIGHT_MOROCCO'                 : 0x045f,
    'TT_MS_LANGID_TAMAZIGHT_MOROCCO_LATIN'           : 0x085f,
    'TT_MS_LANGID_KASHMIRI_PAKISTAN'                 : 0x0460,
    'TT_MS_LANGID_KASHMIRI_SASIA'                    : 0x0860,
    'TT_MS_LANGID_KASHMIRI_INDIA'                    : 0x0860,
    'TT_MS_LANGID_NEPALI_NEPAL'                      : 0x0461,
    'TT_MS_LANGID_NEPALI_INDIA'                      : 0x0861,
    'TT_MS_LANGID_FRISIAN_NETHERLANDS'               : 0x0462,
    'TT_MS_LANGID_PASHTO_AFGHANISTAN'                : 0x0463,
    'TT_MS_LANGID_FILIPINO_PHILIPPINES'              : 0x0464,
    'TT_MS_LANGID_DHIVEHI_MALDIVES'                  : 0x0465,
    'TT_MS_LANGID_DIVEHI_MALDIVES'                   : 0x0465,
    'TT_MS_LANGID_EDO_NIGERIA'                       : 0x0466,
    'TT_MS_LANGID_FULFULDE_NIGERIA'                  : 0x0467,
    'TT_MS_LANGID_HAUSA_NIGERIA'                     : 0x0468,
    'TT_MS_LANGID_IBIBIO_NIGERIA'                    : 0x0469,
    'TT_MS_LANGID_YORUBA_NIGERIA'                    : 0x046a,
    'TT_MS_LANGID_QUECHUA_BOLIVIA'                   : 0x046b,
    'TT_MS_LANGID_QUECHUA_ECUADOR'                   : 0x086b,
    'TT_MS_LANGID_QUECHUA_PERU'                      : 0x0c6b,
    'TT_MS_LANGID_SEPEDI_SOUTH_AFRICA'               : 0x046c,
    'TT_MS_LANGID_SOTHO_SOUTHERN_SOUTH_AFRICA'       : 0x046c,
    'TT_MS_LANGID_IGBO_NIGERIA'                      : 0x0470,
    'TT_MS_LANGID_KANURI_NIGERIA'                    : 0x0471,
    'TT_MS_LANGID_OROMO_ETHIOPIA'                    : 0x0472,
    'TT_MS_LANGID_TIGRIGNA_ETHIOPIA'                 : 0x0473,
    'TT_MS_LANGID_TIGRIGNA_ERYTHREA'                 : 0x0873,
    'TT_MS_LANGID_TIGRIGNA_ERYTREA'                  : 0x0873,
    'TT_MS_LANGID_GUARANI_PARAGUAY'                  : 0x0474,
    'TT_MS_LANGID_HAWAIIAN_UNITED_STATES'            : 0x0475,
    'TT_MS_LANGID_LATIN'                             : 0x0476,
    'TT_MS_LANGID_SOMALI_SOMALIA'                    : 0x0477,
    'TT_MS_LANGID_YI_CHINA'                          : 0x0478,
    'TT_MS_LANGID_PAPIAMENTU_NETHERLANDS_ANTILLES'   : 0x0479,
    'TT_MS_LANGID_UIGHUR_CHINA'                      : 0x0480,
    'TT_MS_LANGID_MAORI_NEW_ZEALAND'                 : 0x0481 }
globals().update(TT_MS_LANGIDS)


# -----------------------------------------------------------------------------
# Possible values of the 'name' identifier field in the name records of the TTF
# 'name' table.  These values are platform independent.
TT_NAME_IDS = {
    'TT_NAME_ID_COPYRIGHT'            :  0,
    'TT_NAME_ID_FONT_FAMILY'          :  1,
    'TT_NAME_ID_FONT_SUBFAMILY'       :  2,
    'TT_NAME_ID_UNIQUE_ID'            :  3,
    'TT_NAME_ID_FULL_NAME'            :  4,
    'TT_NAME_ID_VERSION_STRING'       :  5,
    'TT_NAME_ID_PS_NAME'              :  6,
    'TT_NAME_ID_TRADEMARK'            :  7,

    # the following values are from the OpenType spec
    'TT_NAME_ID_MANUFACTURER'         :  8,
    'TT_NAME_ID_DESIGNER'             :  9,
    'TT_NAME_ID_DESCRIPTION'          : 10,
    'TT_NAME_ID_VENDOR_URL'           : 11,
    'TT_NAME_ID_DESIGNER_URL'         : 12,
    'TT_NAME_ID_LICENSE'              : 13,
    'TT_NAME_ID_LICENSE_URL'          : 14,
    # number 15 is reserved
    'TT_NAME_ID_PREFERRED_FAMILY'     : 16,
    'TT_NAME_ID_PREFERRED_SUBFAMILY'  : 17,
    'TT_NAME_ID_MAC_FULL_NAME'        : 18,

    # The following code is new as of 2000-01-21
    'TT_NAME_ID_SAMPLE_TEXT'          : 19,

    # This is new in OpenType 1.3
    'TT_NAME_ID_CID_FINDFONT_NAME'    : 20,

    # This is new in OpenType 1.5
    'TT_NAME_ID_WWS_FAMILY'           : 21,
    'TT_NAME_ID_WWS_SUBFAMILY'        : 22 }
globals().update(TT_NAME_IDS)


# -----------------------------------------------------------------------------
# A list of bit flags that inform client applications of embedding and
# subsetting restrictions associated with a font.
#
# FT_FSTYPE_INSTALLABLE_EMBEDDING
# Fonts with no fsType bit set may be embedded and permanently installed on
# the remote system by an application.
#
# FT_FSTYPE_RESTRICTED_LICENSE_EMBEDDING
# Fonts that have only this bit set must not be modified, embedded or exchanged
# in any manner without first obtaining permission of the font software
# copyright owner.
#
# FT_FSTYPE_PREVIEW_AND_PRINT_EMBEDDING
# If this bit is set, the font may be embedded and temporarily loaded on the
# remote system. Documents containing Preview & Print fonts must be opened
# 'read-only'; no edits can be applied to the document.
#
# FT_FSTYPE_EDITABLE_EMBEDDING
# If this bit is set, the font may be embedded but must only be installed
# temporarily on other systems. In contrast to Preview & Print fonts,
# documents containing editable fonts may be opened for reading, editing is
# permitted, and changes may be saved.
#
# FT_FSTYPE_NO_SUBSETTING
# If this bit is set, the font may not be subsetted prior to embedding.
#
# FT_FSTYPE_BITMAP_EMBEDDING_ONLY
# If this bit is set, only bitmaps contained in the font may be embedded; no
# outline data may be embedded. If there are no bitmaps available in the font,
# then the font is unembeddable.

FT_FSTYPE_XXX = {'FT_FSTYPE_INSTALLABLE_EMBEDDING'        : 0x0000,
                 'FT_FSTYPE_RESTRICTED_LICENSE_EMBEDDING' : 0x0002,
                 'FT_FSTYPE_PREVIEW_AND_PRINT_EMBEDDING'  : 0x0004,
                 'FT_FSTYPE_EDITABLE_EMBEDDING'           : 0x0008,
                 'FT_FSTYPE_NO_SUBSETTING'                : 0x0100,
                 'FT_FSTYPE_BITMAP_EMBEDDING_ONLY'        : 0x0200,}
globals().update(FT_FSTYPE_XXX)

ft_fstype_installable_embedding  = FT_FSTYPE_INSTALLABLE_EMBEDDING
ft_fstype_restricted_license_embedding = FT_FSTYPE_RESTRICTED_LICENSE_EMBEDDING
ft_fstype_preview_and_print_embedding = FT_FSTYPE_PREVIEW_AND_PRINT_EMBEDDING
ft_fstype_editable_embedding = FT_FSTYPE_EDITABLE_EMBEDDING
ft_fstype_no_subsetting = FT_FSTYPE_NO_SUBSETTING
ft_fstype_bitmap_embedding_only = FT_FSTYPE_BITMAP_EMBEDDING_ONLY



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" Module wavefront

This module produces functionality to read and write wavefront (.OBJ) files.

http://en.wikipedia.org/wiki/Wavefront_.obj_file

The wavefront format is quite powerfull and allows a wide variety of surfaces
to be described.

This implementation does only supports mesh stuff, so no nurbs etc. Further,
material properties are ignored, although this might be implemented later,

The classes are written with compatibility of Python3 in mind.

"""

import visvis as vv
import numpy as np
import time


# todo: have a look at this more efficient approach to reading wavefront
# Alternative method by Nicolas Rougier. Numpy does not seem to have
# a simalar method to write though ...
#
# def load_obj(filename):
#     '''
#     Load vertices and faces from a wavefront .obj file and generate normals.
#     '''
#     data = np.genfromtxt(filename, dtype=[('type', np.character, 1),
#                                           ('points', np.float32, 3)])
#
#     # Get vertices and faces
#     vertices = data['points'][data['type'] == 'v']
#     faces = (data['points'][data['type'] == 'f']-1).astype(np.uint32)
#
#     # Build normals
#     T = vertices[faces]
#     N = np.cross(T[::,1 ]-T[::,0], T[::,2]-T[::,0])
#     L = np.sqrt(N[:,0]**2+N[:,1]**2+N[:,2]**2)
#     N /= L[:, np.newaxis]
#     normals = np.zeros(vertices.shape)
#     normals[faces[:,0]] += N
#     normals[faces[:,1]] += N
#     normals[faces[:,2]] += N
#     L = np.sqrt(normals[:,0]**2+normals[:,1]**2+normals[:,2]**2)
#     normals /= L[:, np.newaxis]
#
#     # Scale vertices such that object is contained in [-1:+1,-1:+1,-1:+1]
#     vmin, vmax =  vertices.min(), vertices.max()
#     vertices = 2*(vertices-vmin)/(vmax-vmin) - 1
#
#     return vertices, normals, faces


class WavefrontReader(object):
    
    def __init__(self, f):
        self._f = f
        
        # Original vertices, normals and texture coords.
        # These are not necessarily of the same length.
        self._v = []
        self._vn = []
        self._vt = []
        
        # Final vertices, normals and texture coords.
        # All three lists are of the same length, as opengl wants it.
        self._vertices = []
        self._normals = []
        self._texcords = []
        
        # The faces, indices to vertex/normal/texcords arrays.
        self._faces = []
        
        # Dictionary to keep track of processed face data, so we can
        # convert the original v/vn/vn to the final vertices/normals/texcords.
        self._facemap = {}
    
    
    @classmethod
    def read(cls, fname, check='ignored'):
        """ read(fname)
        
        This classmethod is the entry point for reading OBJ files.
        
        Parameters
        ----------
        fname : string
            The name of the file to read.
        
        """
        
        t0 = time.time()  # noqa
        
        # Open file
        f = open(fname, 'rb')
        try:
            reader = WavefrontReader(f)
            while True:
                reader.readLine()
        except EOFError:
            pass
        finally:
            f.close()
        
        # Done
        mesh = reader.finish()
        #print('reading mesh took ' + str(time.time()-t0) + ' seconds')
        return mesh
    
    
    def readLine(self):
        """ The method that reads a line and processes it.
        """
        
        # Read line
        line = self._f.readline().decode('ascii', 'ignore')
        if not line:
            raise EOFError()
        line = line.strip()
        
        if line.startswith('v '):
            #self._vertices.append( *self.readTuple(line) )
            self._v.append( self.readTuple(line) )
        elif line.startswith('vt '):
            self._vt.append( self.readTuple(line, 3) )
        elif line.startswith('vn '):
            self._vn.append( self.readTuple(line) )
        elif line.startswith('f '):
            self._faces.append( self.readFace(line) )
        elif line.startswith('#'):
            pass # Comment
        elif line.startswith('mtllib '):
            print('Notice reading .OBJ: material properties are ignored.')
        elif line.startswith('g ') or line.startswith('s '):
            pass # Ignore groups and smoothing groups
        elif line.startswith('o '):
            pass # Ignore object names
        elif line.startswith('usemtl '):
            pass # Ignore material
        elif not line.strip():
            pass
        else:
            print('Notice reading .OBJ: ignoring %s command.' % line.strip())
    
    
    def readTuple(self, line, n=3):
        """ Reads a tuple of numbers. e.g. vertices, normals or teture coords.
        """
        numbers = [num for num in line.split(' ') if num]
        return [float(num) for num in numbers[1:n+1]]
    
    
    def readFace(self, line):
        """ Each face consists of three or more sets of indices. Each set
        consists of 1, 2 or 3 indices to vertices/normals/texcords.
        """
        
        # Get parts (skip first)
        indexSets = [num for num in line.split(' ') if num][1:]
        
        final_face = []
        for indexSet in indexSets:
            
            # Did we see this exact index earlier? If so, it's easy
            final_index = self._facemap.get(indexSet)
            if final_index is not None:
                final_face.append(final_index)
                continue
            
            # If not, we need to sync the vertices/normals/texcords ...
            
            # Get and store final index
            final_index = len(self._vertices)
            final_face.append(final_index)
            self._facemap[indexSet] = final_index
            
            # What indices were given?
            indices = [i for i in indexSet.split('/')]
            
            # Store new set of vertex/normal/texcords.
            # If there is a single face that does not specify the texcord
            # index, the texcords are ignored. Likewise for the normals.
            if True:
                vertex_index = self._absint(indices[0], len(self._v))
                self._vertices.append( self._v[vertex_index] )
            if self._texcords is not None:
                if len(indices) > 1 and indices[1]:
                    texcord_index = self._absint(indices[1], len(self._vt))
                    self._texcords.append( self._vt[texcord_index] )
                else:
                    if self._texcords:
                        print('Warning reading OBJ: ignoring texture coordinates because '
                              'it is not specified for all faces.')
                    self._texcords = None
            if self._normals is not None:
                if len(indices) > 2 and indices[2]:
                    normal_index = self._absint(indices[2], len(self._vn))
                    self._normals.append( self._vn[normal_index] )
                else:
                    if self._normals:
                        print('Warning reading OBJ: ignoring normals because it is not '
                              'specified for all faces.')
                    self._normals = None
        
        # Check face
        if self._faces and len(self._faces[0]) != len(final_face):
            raise RuntimeError('Visvis requires that all faces are either triangles or quads.')
        
        # Done
        return final_face
    
    
    def _absint(self, i, ref):
        i = int(i)
        if i>0 :
            return i-1
        else:
            return ref+i
    
    
    def finish(self):
        """ Converts gathere lists to numpy arrays and creates
        BaseMesh instance.
        """
        if True:
            self._vertices = np.array(self._vertices, 'float32')
        if self._normals:
            self._normals = np.array(self._normals, 'float32')
        else:
            self._normals = None
        if self._texcords:
            self._texcords = np.array(self._texcords, 'float32')
        else:
            self._texcords = None
        if self._faces:
            self._faces = np.array(self._faces, 'uint32')
        else:
            # Use vertices only
            self._vertices = np.array(self._v, 'float32')
            self._faces = None
        
        return vv.BaseMesh(self._vertices, self._faces, self._normals, self._texcords)
    


class WavefrontWriter(object):
    
    def __init__(self, f):
        self._f = f
    
    
    @classmethod
    def write(cls, fname, mesh, name='', bin='ignored'):
        """ write(fname, mesh, name='')
        
        This classmethod is the entry point for writing mesh data to OBJ.
        
        Parameters
        ----------
        fname : string
            The filename to write to.
        mesh : Mesh data
            Can be vv.BaseMesh, vv.Pointset, or np.ndarray.
        name : string
            The name of the object (e.g. 'teapot')
        
        """
        
        # Open file
        f = open(fname, 'wb')
        try:
            writer = WavefrontWriter(f)
            writer.writeMesh(mesh, name)
        except EOFError:
            pass
        finally:
            f.close()
    
    
    def writeLine(self, text):
        """ Simple writeLine function to write a line of code to the file.
        The encoding is done here, and a newline character is added.
        """
        text += '\n'
        self._f.write(text.encode('ascii'))
    
    
    def writeTuple(self, val, what):
        """ Writes a tuple of numbers (on one line).
        """
        # Limit to three values. so RGBA data drops the alpha channel
        # Format can handle up to 3 texcords
        val = val[:3]
        # Make string
        val = ' '.join([str(v) for v in val])
        # Write line
        self.writeLine('%s %s' % (what, val))

    
    def writeFace(self, val, what='f'):
        """ Write the face info to the net line.
        """
        # OBJ counts from 1
        val = [v+1 for v in val]
        # Make string
        if self._hasValues and self._hasNormals:
            val = ' '.join(['%i/%i/%i' % (v,v,v) for v in val])
        elif self._hasNormals:
            val = ' '.join(['%i//%i' % (v,v) for v in val])
        elif self._hasValues:
            val = ' '.join(['%i/%i' % (v,v) for v in val])
        else:
            val = ' '.join(['%i' % v for v in val])
        # Write line
        self.writeLine('%s %s' % (what, val))
    
    
    def writeMesh(self, mesh, name=''):
        """ Write the given mesh instance.
        """
        
        # Store properties
        self._hasNormals = mesh._normals is not None
        self._hasValues = mesh._values is not None
        self._hasFaces = mesh._faces is not None
        
        # Get faces and number of vertices
        faces = mesh._GetFaces()
        N = mesh._vertices.shape[0]
        
        # Get string with stats
        stats = []
        stats.append('%i vertices' % N)
        if self._hasValues:
            stats.append('%i texcords' % N)
        else:
            stats.append('no texcords')
        if self._hasNormals:
            stats.append('%i normals' % N)
        else:
            stats.append('no normals')
        stats.append('%i faces' % faces.shape[0])
        
        
        # Write header
        self.writeLine('# Wavefront OBJ file')
        self.writeLine('# Created by visvis (a Python visualization toolkit).')
        self.writeLine('#')
        if name:
            self.writeLine('# object %s' % name)
        else:
            self.writeLine('# unnamed object')
        self.writeLine('# %s' % ', '.join(stats))
        self.writeLine('')
        
        # Write data
        if True:
            for i in range(N):
                self.writeTuple(mesh._vertices[i], 'v')
        if self._hasNormals:
            for i in range(N):
                self.writeTuple(mesh._normals[i], 'vn')
        if self._hasValues:
            for i in range(N):
                self.writeTuple(mesh._values[i], 'vt')
        if True:
            for i in range(faces.shape[0]):
                self.writeFace(faces[i])



#!/usr/bin/env python
""" This example illustrates how to create an isocontour from an image
and display it over the contour. The contour consists of a pointset
in which each 2 subsequent points describe a linepiece.

This code relies on scikit-image. For a possibly more useful representation
of the contour, see skimage.measure.find_contours.
"""

import visvis as vv

im = vv.imread('imageio:chelsea.png')[:,:,1]
pp = vv.isocontour(im)

vv.figure(1)
vv.clf()
vv.imshow(im)
vv.plot(pp, ls='+', lw=2)

vv.title('Isocontour')
app = vv.use()
app.Run()



#!/usr/bin/env python

""" This example illustrates using transparant textures. This can be used,
for example to overlay a segmentation result over an original image.

Note that the alpha value is always asumed between 0 and 1, so a
transparant texture should always be of float type.
"""

import visvis as vv
import numpy as np
app = vv.use()

# Lena is our original image
vv.clf()
im = vv.imread('astronaut.png')

# Find the regions where there's relatively much blue
mask = (im[:,:,0] < 200) & (im[:,:,2]>0.7*im[:,:,0])

# Create an RGBA texture and fill in the found region in blue
mask2 = np.zeros(mask.shape+(4,), dtype=np.float32)
mask2[:,:,2] = mask
mask2[:,:,3] = mask * 0.5

# Add a black, green, red, and yellow region in the corner
mask2[100:200,:200,0] = 1.0
mask2[:200,100:200,1] = 1.0
mask2[:200,:200,3] = 0.5
  
# Show image and mask
t1=vv.imshow(im)
t2=vv.imshow(mask2)

app.Run()



#!/usr/bin/env python

import visvis as vv
import numpy as np

# Define angles
angs = 0.1 + np.linspace(-90, 90, 181)  # 0.1+ get rid of singularity
angsRads = np.pi * angs / 180.0

# Define magnitude
mag = 10 * np.log10(np.abs(np.sin(10 * angsRads) / angsRads)) + angsRads
mag = mag - np.max(mag)

# Plot two versions of the signal, one rotated by 20 degrees
vv.polarplot( angs, mag, lc='b')
vv.polarplot(angs+20, mag, lc='r', lw=2)

# Set axis properties
ax = vv.gca()

ax.axis.angularRefPos = 90  # 0 deg points up
ax.axis.isCW = False  # changes angular sense (azimuth instead of phi)
ax.axis.SetLimits( rangeTheta=0, rangeR=vv.Range(-40, 5))

ax.axis.xLabel = 'degrees'
ax.axis.yLabel = 'dB'
ax.axis.showGrid = True

# Show some messages
print('drag mouse left button left-right to rotate plot.')
print('shift-drag mouse left button up-down to change lower')
print('radial limit. Drag mouse right button up-down to rescale')
print('radial axis while keeping lower radial limit fixed')

# Run mainloop
app = vv.use()
app.Run()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import visvis as vv

def gca():
    """ gca()
    
    Get the current axes in the current figure. If there is
    no axes, an Axes instance is created. To make an axes current,
    use Axes.MakeCurrent().
    
    See also gcf(), Figure.MakeCurrent(), Figure.currentAxes
    
    """
    f = vv.gcf()
    a = f.currentAxes
    if not a:
        # create axes
        a = vv.Axes(f)
        #a.position = 2, 2, -4, -4
        a.position = 10, 10, -20, -20
    return a
    
    
if __name__ == '__main__':
    a = vv.gca()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import visvis as vv

def title(text, axes=None):
    """ title(text, axes=None)
    
    Show title above axes. Remove title by suplying an empty string.
    
    Parameters
    ----------
    text : string
        The text to display.
    axes : Axes instance
        Display the image in this axes, or the current axes if not given.
    
    """
    
    if axes is None:
        axes = vv.gca()
    
    # seek Title object
    for child in axes.children:
        if isinstance(child, vv.Title):
            ob = child
            ob.text = text
            break
    else:
        ob = vv.Title(axes, text)
    
    # destroy if no text, return object otherwise
    if not text:
        ob.Destroy()
        return None
    else:
        return ob


if __name__=='__main__':
    a = vv.gca()
    vv.title('test title')
